import copy
from itertools import chain
from pathlib import Path
from tempfile import TemporaryFile
from typing import Dict, List, Optional, Union

import numpy as np
import pandas as pd
from stl import mesh

from pyskindose.plotting.create_ploty_ijk_indices import (
    _create_plotly_ijk_indices_for_cuboid_objects,
)
from pyskindose.settings.phantom_dimensions import PhantomDimensions

# valid phantom types
VALID_PHANTOM_MODELS = ["plane", "cylinder", "human", "table", "pad"]


class Phantom:
    """Create and handle phantoms for patient, support table and pad.

    This class creates a phantom of any of the types specified in
    VALID_PHANTOM_MODELS (plane, cylinder or human to represent the patient,
    as well as patient support table and pad). The patient phantoms consists of
    a number of skin cells where the skin dose can be calculated.

    Attributes
    ----------
    phantom_model : str
        Type of phantom, i.e. "plane", "cylinder", "human", "table" or "pad"
    r : np.ndarray
        n*3 array where n are the number of phantom skin cells. Each row
        contains the xyz coordinate of one of the phantom skin cells
    ijk : np.array
        A matrix containing vertex indices. This is required in order to
        plot the phantom using plotly Mesh3D. For more info, see "i", "j", and
        "k" at https://plot.ly/python/reference/#mesh3d
    dose : np.array
        An empty 1d array to store skin dose calculation for each of the n
        phantom cells. Only for patient phantom types (plane, cylinder, human)
    n : np.array
        normal vectors to each of the n phantom skin cells. (only for 3D
        patient phantoms, i.e. "cylinder" and "human")
    r_ref : np.array
        Empty array to store of reference position of the phantom cells after
        the phantom has been aligned in the geometry with the
        position_patient_phantom_on_table function in geom_calc.py
    table_length : float
        length of patient support table. The is needed for all phantom object
        to select correct rotation origin for At1, At2, and At3.


    """

    def __init__(
        self,
        phantom_model: str,
        phantom_dim: PhantomDimensions,
        human_mesh: Optional[Union[str, tuple[str, mesh.Mesh]]] = None,
    ):
        """Create the phantom of choice.

        Parameters
        ----------
        phantom_model : str
            Type of phantom to create. Valid selections are 'plane',
            'cylinder', 'human', "table", and "pad".
        phantom_dim : PhantomDimensions
            instance of class PhantomDimensions containing dimensions for
            all phantoms models except human phantoms: Length, width, radius,
            thickness etc.
        human_mesh : str | tuple[str, mes.Mesh | temp_file], optional
            Choose which human mesh phantom to use. Valid selection are names
            of the *.stl-files in the phantom_data folder or a custom phantom
            sent in as either a mesh object or a svg given as a temp_file object
            (The default is none).

        Raises
        ------
        ValueError
            Raises value error if unsupported phantom type are selected,
            or if phantom_model='human' selected, without specifying
            human_mesh

        """
        self.phantom_model = phantom_model.lower()
        # Raise error if invalid phantom model selected
        if self.phantom_model not in VALID_PHANTOM_MODELS:
            raise ValueError(f"Unknown phantom model selected. Valid type:" f"{'.'.join(VALID_PHANTOM_MODELS)}")

        self.human_model = None

        self.r_ref: np.array

        # Save table length for all phantom in order to choose correct rotation
        # origin when applying At1, At2, and At3
        self.table_length = phantom_dim.table_length

        # creates a plane phantom (2D grid)
        if phantom_model == "plane":

            # Use a dense grid if specified by user
            if phantom_dim.plane_resolution.lower() == "dense":
                res_length = res_width = 2.0

            elif phantom_dim.plane_resolution.lower() == "sparse":
                res_length = res_width = 1.0

            # Linearly spaced points along the longitudinal direction
            x = np.linspace(
                -phantom_dim.plane_width / 2, +phantom_dim.plane_width / 2, int(res_width * phantom_dim.plane_width + 1)
            )
            # Linearly spaced points along the lateral direction
            z = np.linspace(0, -phantom_dim.plane_length, int(res_length * phantom_dim.plane_length))

            # Create phantom in form of rectangular grid
            x_plane, z_plane = np.meshgrid(x, z)

            t = phantom_dim.plane_width

            # Create index vectors for plotly mesh3d plotting
            i2: List[int] = []
            i1 = j1 = k1 = i2

            for i in range(len(x) - 1):
                for j in range(len(z) - 1):
                    i1 = i1 + [j * len(x) + i]
                    j1 = j1 + [j * len(x) + i + 1]
                    k1 = k1 + [j * len(x) + i + len(x)]
                    i2 = i2 + [j * len(x) + i + len(x) + 1]

            self.r = np.column_stack((x_plane.ravel(), np.zeros(len(x_plane.ravel())), z_plane.ravel()))

            self.ijk = np.column_stack((i1 + i2, j1 + k1, k1 + j1))
            self.dose = np.zeros(len(self.r))

        # creates a cylinder phantom (elliptic)
        elif phantom_model == "cylinder":

            # Use a dense grid if specified by user
            if phantom_dim.cylinder_resolution.lower() == "dense":
                res_length = 4
                res_width = 0.05

            elif phantom_dim.cylinder_resolution.lower() == "sparse":
                res_length = 1.0
                res_width = 0.1

            # Creates linearly spaced points along an ellipse
            #  in the lateral direction
            t = np.arange(0 * np.pi, 2 * np.pi, res_width)
            x = (phantom_dim.cylinder_radii_a * np.cos(t)).tolist()
            y = (phantom_dim.cylinder_radii_b * np.sin(t)).tolist()

            # calculate normal vectors of a cylinder (pointing outwards)
            nx = np.cos(t) / (np.sqrt(np.square(np.cos(t) + 4 * np.square(np.sin(t)))))

            nz = np.zeros(len(t))

            ny = 2 * np.sin(t) / (np.sqrt(np.square(np.cos(t) + 4 * np.square(np.sin(t)))))

            nx = nx.tolist()
            ny = ny.tolist()
            nz = nz.tolist()

            n = [[nx[ind], ny[ind], nz[ind]] for ind in range(len(t))]

            # Store the  coordinates of the cylinder phantom, extended to span the entire length of the phantom, thus
            # creating an elliptical cylinder
            tmp_len = int(res_length) * (phantom_dim.cylinder_length + 2)
            output: dict = {
                "n": n * tmp_len,
                "x": x * tmp_len,
                "y": [el - phantom_dim.cylinder_radii_b for el in (y * tmp_len)],
                "z": list(chain(*[[-1 / res_length * ind] * len(x) for ind in range(tmp_len)])),
            }

            # Create index vectors for plotly mesh3d plotting
            i1 = list(range(0, len(output["x"]) - len(t)))
            j1 = list(range(1, len(output["x"]) - len(t) + 1))
            k1 = list(range(len(t), len(output["x"])))
            i2 = list(range(0, len(output["x"]) - len(t)))
            k2 = list(range(len(t) - 1, len(output["x"]) - 1))
            j2 = list(range(len(t), len(output["x"])))

            self.r = np.column_stack((output["x"], output["y"], output["z"]))
            self.ijk = np.column_stack((i1 + i2, j1 + j2, k1 + k2))
            self.dose = np.zeros(len(self.r))
            self.n = np.asarray(output["n"])

        # creates a human phantom
        elif phantom_model == "human":

            if human_mesh is None:
                raise ValueError("Human model needs to be specified for" 'phantom_model = "human"')

            if isinstance(human_mesh, str):
                # load selected phantom model from binary .stl file
                self.human_model = human_mesh
                phantom_path = Path(__file__).parent / f"phantom_data/{human_mesh}.stl"
                phantom_mesh = mesh.Mesh.from_file(str(phantom_path.absolute()))
            elif isinstance(human_mesh, tuple):
                self.human_model, phantom_mesh = self._get_phantom_mesh_from_tuple(human_mesh)
            else:
                raise ValueError("No human model specified while 'phantom_model' is 'human'")

            r = phantom_mesh.vectors
            n = phantom_mesh.normals

            self.r = np.asarray([el for el_list in r for el in el_list])
            self.n = np.asarray([x for pair in zip(n, n, n) for x in pair])

            # Create index vectors for plotly mesh3d plotting
            self.ijk = np.column_stack(
                (np.arange(0, len(self.r) - 3, 3), np.arange(1, len(self.r) - 2, 3), np.arange(2, len(self.r) - 1, 3))
            )
            self.dose = np.zeros(len(self.r))

        # Creates the vertices of the patient support table
        elif phantom_model == "table":
            # Longitudinal position of the vertices
            x_tab = [index * phantom_dim.table_width for index in [+0.5, +0.5, -0.5, -0.5, +0.5, +0.5, -0.5, -0.5]]

            # Vertical position of the vertices
            y_tab = [index * phantom_dim.table_thickness for index in [0, 0, 0, 0, +1, +1, +1, +1]]

            # Lateral position of the vertices
            z_tab = [index * phantom_dim.table_length for index in [0, -1, -1, 0, 0, -1, -1, 0]]

            # Create index vectors for plotly mesh3d plotting
            i_tab, j_tab, k_tab = _create_plotly_ijk_indices_for_cuboid_objects()

            self.r = np.column_stack((x_tab, y_tab, z_tab))
            self.ijk = np.column_stack((i_tab, j_tab, k_tab))

        # Creates the vertices of the patient support table
        elif phantom_model == "pad":

            # Longitudinal position of the vertices
            x_pad = [index * phantom_dim.pad_width for index in [+0.5, +0.5, -0.5, -0.5, +0.5, +0.5, -0.5, -0.5]]

            # Vertical position of the vertices
            y_pad = [index * phantom_dim.pad_thickness for index in [0, 0, 0, 0, -1, -1, -1, -1]]

            # Lateral position of the vertices
            z_pad = [index * phantom_dim.pad_length for index in [0, -1, -1, 0, 0, -1, -1, 0]]

            # Create index vectors for plotly mesh3d plotting
            i_pad, j_pad, k_pad = _create_plotly_ijk_indices_for_cuboid_objects()

            self.r = np.column_stack((x_pad, y_pad, z_pad))
            self.ijk = np.column_stack((i_pad, j_pad, k_pad))

    @staticmethod
    def _get_phantom_mesh_from_tuple(
        phantom_mesh_tuple: tuple[str, Union[mesh.Mesh, TemporaryFile, str]]
    ) -> tuple[str, mesh.Mesh]:
        if not isinstance(phantom_mesh_tuple[0], str):
            raise TypeError(
                "If human_mesh is specified as a tuple, the first element must be the phantom name as a string"
            )

        if isinstance(phantom_mesh_tuple[1], mesh.Mesh):
            return phantom_mesh_tuple

        return phantom_mesh_tuple[0], mesh.Mesh.from_file(phantom_mesh_tuple[1])

    def rotate(self, angles: List[int]) -> None:
        """Rotate the phantom about the angles specified in rotation.

        Parameters
        ----------
        angles: List[int]
            list of angles in degrees the phantom should be rotated about,
            given as [x_rot: <int>, y_rot: <int>, z_rot: <int>]. E.g.
            rotation = [0, 90, 0] will rotate the phantom 90 degrees about the
            y-axis.

        """
        # convert degrees to radians
        angles = np.deg2rad(angles)

        x_rot = angles[0]
        y_rot = angles[1]
        z_rot = angles[2]

        # Define rotation matricies about the x, y and z axis
        Rx = np.array([[+1, +0, +0], [+0, +np.cos(x_rot), -np.sin(x_rot)], [+0, +np.sin(x_rot), +np.cos(x_rot)]])
        Ry = np.array([[+np.cos(y_rot), +0, +np.sin(y_rot)], [+0, +1, +0], [-np.sin(y_rot), +0, +np.cos(y_rot)]])
        Rz = np.array([[+np.cos(z_rot), -np.sin(z_rot), +0], [+np.sin(z_rot), +np.cos(z_rot), +0], [+0, +0, +1]])

        # Rotate position vectors to the phantom cells

        self.r = np.matmul(Rx, np.matmul(Ry, np.matmul(Rz, self.r.T))).T

        if self.phantom_model in ["cylinder", "human"]:

            self.n = np.matmul(Rx, np.matmul(Ry, np.matmul(Rz, self.n.T))).T

    def translate(self, dr: List[int]) -> None:
        """Translate the phantom in the x, y or z direction.

        Parameters
        ----------
        dr : List[int]
            list of distances the phantom should be translated, given in cm.
            Specified as dr = [dx: <int>, dy: <int>, dz: <int>]. E.g.
            dr = [0, 0, 10] will translate the phantom 10 cm in the z direction

        """
        self.r[:, 0] += dr[0]
        self.r[:, 1] += dr[1]
        self.r[:, 2] += dr[2]

    def save_position(self) -> None:
        """Store a reference position of the phantom.

        This function is supposed to be used to store the patient fixation
        conducted in the function position_patient_phantom_on_table

        """
        r_ref = copy.copy(self.r)
        self.r_ref = r_ref

    def position(self, data_norm: pd.DataFrame, event: int) -> None:
        """Position the phantom for a event by adding RDSR table displacement.

        Positions the phantom from reference position to actual position
        according to the table displacement info in data_norm.

        Parameters
        ----------
        data_norm : pd.DataFrame
            Table containing dicom RDSR information from each irradiation event
            See rdsr_normalizer.py for more information.
        event : int
            Irradiation event index

        """
        self.r = copy.copy(self.r_ref)

        # displace phantom to table rotation center
        self.r[:, 2] += self.table_length / 2

        # Apply table rotation
        self.r = np.matmul(
            data_norm.Rz[event], np.matmul(data_norm.Ry[event], np.matmul(data_norm.Rx[event], self.r.T))
        ).T

        # Replace phantom back to starting position
        self.r[:, 2] -= self.table_length / 2

        # Apply phantom translation
        t = np.array([data_norm.Tx[event], data_norm.Ty[event], data_norm.Tz[event]])

        self.r = self.r + t



from typing import Dict


def get_camera_view() -> Dict[str, Dict[str, float]]:
    return dict(
        up=dict(x=0, y=-1, z=0),
        center=dict(x=0, y=0, z=0),
        eye=dict(x=-1.3, y=-1.3, z=0.7),
    )



from typing import Any, Dict, Union

import pandas as pd

from pyskindose import constants as c
from pyskindose.calculate_dose.calculate_dose import calculate_dose
from pyskindose.format_export_data import (
    PySkinDoseOutput,
    format_analysis_result_for_export,
)
from pyskindose.helpers.calculate_rotation_matrices import calculate_rotation_matrices
from pyskindose.phantom_class import Phantom
from pyskindose.plotting.create_dose_map_plot import create_dose_map_plot
from pyskindose.plotting.create_geometry_plot import create_geometry_plot
from pyskindose.settings import PyskindoseSettings, initialize_settings


def analyze_data(
    normalized_data: pd.DataFrame,
    settings: Union[str, dict, PyskindoseSettings],
) -> Union[Dict[str, Any], str, PySkinDoseOutput]:
    """Analyze data och settings, and runs PySkinDose in desired mode.

    Parameters
    ----------
    normalized_data : pd.DataFrame
        RDSR data, normalized for compliance with PySkinDose.
    settings : Union[str, dict, PyskindoseSettings]
        Settings class for PySkinDose

    Returns
    -------
    Dict[str, Any]
        output dictionary containing calculation specifics such as dose map, correction
        factors, etc.

    """
    settings = initialize_settings(settings)

    if settings.output_format not in c.RUN_ARGUMENTS_VALID_OUTPUT_FORMATS:
        raise ValueError(
            f"Invalid output format specified. Must be one of {'.'.join(c.RUN_ARGUMENTS_VALID_OUTPUT_FORMATS)}"
        )

    # create table, pad and patient phantoms.
    table = Phantom(phantom_model=c.PHANTOM_MODEL_TABLE, phantom_dim=settings.phantom.dimension)

    pad = Phantom(phantom_model=c.PHANTOM_MODEL_PAD, phantom_dim=settings.phantom.dimension)

    normalized_data = calculate_rotation_matrices(normalized_data)

    create_geometry_plot(normalized_data=normalized_data, table=table, pad=pad, settings=settings)

    patient, output = calculate_dose(normalized_data=normalized_data, settings=settings, table=table, pad=pad)

    if settings.output_format in [c.RUN_ARGUMENTS_OUTPUT_DICT, c.RUN_ARGUMENTS_OUTPUT_JSON]:
        pyskindose_output: Union[PySkinDoseOutput, dict[str, Any], str] = format_analysis_result_for_export(
            output, patient=patient, table=table, pad=pad, data_norm=normalized_data, settings=settings
        )

        return pyskindose_output

    create_dose_map_plot(
        patient=patient,
        settings=settings,
        dose_map=(
            output[c.OUTPUT_KEY_DOSE_MAP] if settings.mode in (c.MODE_CALCULATE_DOSE, c.MODE_PLOT_DOSEMAP) else None
        ),
    )

    if settings.output_format == c.RUN_ARGUMENTS_OUTPUT_HTML:
        return output



import json
from dataclasses import dataclass
from typing import Any, Dict, Union

import numpy as np
import pandas as pd

from pyskindose import Beam, Phantom
from pyskindose.constants import (
    KEY_NORMALIZATION_AIR_KERMA,
    OUTPUT_KEY_CORRECTION_BACK_SCATTER,
    OUTPUT_KEY_CORRECTION_INVERSE_SQUARE_LAW,
    OUTPUT_KEY_CORRECTION_MEDIUM,
    OUTPUT_KEY_CORRECTION_TABLE,
    OUTPUT_KEY_DOSE_MAP,
    OUTPUT_KEY_HITS,
    PHANTOM_MODEL_HUMAN,
    PLOT_TRACE_ORDER_BEAM_WIREFRAME,
    PLOT_TRACE_ORDER_DETECTOR_WIREFRAME,
    PLOT_TRACE_ORDER_PHANTOM_WIREFRAME,
    RUN_ARGUMENTS_OUTPUT_DICT,
    RUN_ARGUMENTS_OUTPUT_JSON,
)
from pyskindose.settings import PyskindoseSettings


@dataclass
class Position:
    """Create and handle the x, y, and z-positions of, e.g., a phantom. When used for a phantom, each combination of an
    element of the same index in the x, y, and z-list represents the position of one skin cell.

    Attributes
    ----------
    x : list[float]
        A list of the x-positions, e.g., representing the x-position of a phantom's skin cells
    y : list[float]
        A list of the y-positions, e.g., representing the y-position of a phantom's skin cells
    z : list[float]
        A list of the z-positions, e.g., representing the x-position of a phantom's skin cells
    """

    x: list[float]
    y: list[float]
    z: list[float]

    def to_dict(self):
        return {
            "x": [float(el) for el in self.x],
            "y": [float(el) for el in self.y],
            "z": [float(el) for el in self.z],
        }


@dataclass
class VertexIndices:
    """Create and handle the x, y, and z-positions of, e.g., a phantom. When used for a phantom, each combination of an
    element of the same index in the x, y, and z-list represents the position of one skin cell.

    Attributes
    ----------
    i : list[float]
        A list of the i-vertex indies, e.g., representing the i-vertex indices of a phantom's skin cells
    j : list[float]
        A list of the j-vertex indices, e.g., representing the j--vertex indices of a phantom's skin cells
    k : list[float]
        A list of the k-vertex indices, e.g., representing the k-vertex indices of a phantom's skin cells
    """

    i: list[float]
    j: list[float]
    k: list[float]

    def to_dict(self):
        return {
            "i": [float(el) for el in self.i],
            "j": [float(el) for el in self.j],
            "k": [float(el) for el in self.k],
        }


class HumanPhantomOutput:
    """Create and handle a patient phantom data for output into a dict or JSON-string.

    Attributes
    ----------
    human_model : str
        The name of the human model used in the PySkinDose calculation
    phantom_skin_cells : Position
        The positions of all the phantom skin cells
    triangle_vertex_indices : VertexIndices
        The vertex indices of all the phantom skin cells
    r_ref : np.array
        The reference position of the phantom cells after the phantom has been aligned
        in the geometry with the position_patient_phantom_on_table function in geom_calc.py
    """

    def __init__(self, phantom: Phantom):
        self.human_model = phantom.human_model
        self.phantom_skin_cells = Position(
            x=phantom.r[:, 0].tolist(), y=phantom.r[:, 1].tolist(), z=phantom.r[:, 2].tolist()
        )
        self.triangle_vertex_indices = VertexIndices(
            i=phantom.ijk[:, 0].tolist(), j=phantom.ijk[:, 1].tolist(), k=phantom.ijk[:, 2].tolist()
        )
        self.r_ref = phantom.r_ref

    def to_dict(self) -> dict:
        return {
            "human_phantom": self.human_model,
            "r_ref": self.r_ref.tolist(),
            "patient_skin_cells": self.phantom_skin_cells.to_dict(),
            "triangle_vertex_indices": self.triangle_vertex_indices.to_dict(),
        }

    def to_json(self) -> str:
        return json.dumps(self.to_dict())


class NonHumanPhantomOutput(HumanPhantomOutput):
    """Create and handle non-human phantoms, that is all phantoms that are not using a human model from, e.g., an
    stl-file.

    Attributes
    ----------
    phantom_skin_cells : Position
        The positions of all the phantom skin cells
    triangle_vertex_indices : VertexIndices
        The vertex indices of all the phantom skin cells
    """

    def __init__(self, phantom: Phantom):
        super().__init__(phantom)
        self.human_model = None


class EventOutput:
    """Create and handle the data specifying an irradiation event, e.g., the positioning of the patient, table, pad, and
    beam.

    Attributes
    ----------
    events: int
        The number of events included in the PySkinDose calculation
    rotation : dict[str, list[float]]
        The x, y, and z rotation for each event
    translation : dict[str, list[float]]
        The x, y, and z translation for each event
    beam_positions : list[Position]
        The position of the beam for each event
    beam_vertex_indices : list[VertexIndices]
        The vertex indices of the beam for each event
    detector_positions : list[Position]
        The position of the detector for each event
    detector_vertex_indices : list[VertexIndices]
        The vertex indices of the detector for each event
    phantom_object_trace_order : list[int]
        The trace order to for the phantom object when creating plotly plots
    beam_wireframe_trace_order : list[int]
        The trace order to for the beam wireframes when creating plotly plots
    detector_wireframe_trace_order : list[int]
        The trace order to for the detector object when creating plotly plots
    """

    def __init__(self, data_norm: pd.DataFrame):
        self.events = len(data_norm)

        self.rotation = {
            "x": data_norm["Rx"].tolist(),
            "y": data_norm["Ry"].tolist(),
            "z": data_norm["Rz"].tolist(),
        }
        self.translation = {
            "x": data_norm.Tx.tolist(),
            "y": data_norm.Ty.tolist(),
            "z": data_norm.Tz.tolist(),
        }
        self.beam_positions, self.beam_vertex_indices, self.detector_positions, self.detector_vertex_indices = zip(
            *[self._extract_beam_data_list(data_norm=data_norm, event=event) for event in range(len(data_norm))]
        )
        self.phantom_object_trace_order = PLOT_TRACE_ORDER_PHANTOM_WIREFRAME
        self.beam_wireframe_trace_order = PLOT_TRACE_ORDER_BEAM_WIREFRAME
        self.detector_wireframe_trace_order = PLOT_TRACE_ORDER_DETECTOR_WIREFRAME
        (
            self.setup_beam_positions,
            self.setup_beam_vertex_indices,
            self.setup_detector_positions,
            self.setup_detector_vertex_indices,
        ) = self._extract_beam_data_list(data_norm=data_norm, event=0, setup=True)

    def _extract_position_list(self, phantom: Phantom, data_norm: pd.DataFrame) -> list[Position]:
        return [
            self._get_position_dict(phantom=phantom, data_norm=data_norm, event=ind) for ind in range(len(data_norm))
        ]

    @staticmethod
    def _get_position_dict(phantom: Phantom, data_norm: pd.DataFrame, event: int) -> Position:
        phantom.position(data_norm=data_norm, event=event)
        return Position(
            x=phantom.r[:, 0].tolist(),
            y=phantom.r[:, 1].tolist(),
            z=phantom.r[:, 2].tolist(),
        )

    @staticmethod
    def _extract_beam_data_list(
        data_norm: pd.DataFrame, event: int, setup: bool = False
    ) -> tuple[Position, VertexIndices, Position, VertexIndices]:
        beam = Beam(data_norm, event=event, plot_setup=setup)
        beam_position = Position(x=beam.r[:, 0].tolist(), y=beam.r[:, 1].tolist(), z=beam.r[:, 2].tolist())
        beam_vertex_indices = VertexIndices(
            i=beam.ijk[:, 0].tolist(), j=beam.ijk[:, 1].tolist(), k=beam.ijk[:, 2].tolist()
        )
        detector_position = Position(
            x=beam.det_r[:, 0].tolist(), y=beam.det_r[:, 1].tolist(), z=beam.det_r[:, 2].tolist()
        )
        detector_vertex_indices = VertexIndices(
            i=beam.det_ijk[:, 0].tolist(), j=beam.det_ijk[:, 1].tolist(), k=beam.det_ijk[:, 2].tolist()
        )

        return beam_position, beam_vertex_indices, detector_position, detector_vertex_indices

    def to_dict(self):
        return {
            "number_of_events": self.events,
            "rotation": self.rotation,
            "translation": self.translation,
            "phantom_object_trace_order": self.phantom_object_trace_order,
            "beam": {
                "positions": [pos.to_dict() for pos in self.beam_positions],
                "vertex_indices": [pos.to_dict() for pos in self.beam_vertex_indices],
                "trace_order": self.beam_wireframe_trace_order,
                "setup": {
                    "position": self.setup_beam_positions.to_dict(),
                    "vertex_indices": self.setup_beam_vertex_indices.to_dict(),
                },
            },
            "detector": {
                "positions": [pos.to_dict() for pos in self.detector_positions],
                "vertex_indices": [pos.to_dict() for pos in self.detector_vertex_indices],
                "trace_order": self.detector_wireframe_trace_order,
                "setup": {
                    "position": self.setup_detector_positions.to_dict(),
                    "vertex_indices": self.setup_detector_vertex_indices.to_dict(),
                },
            },
        }


class PySkinDoseOutput:
    """A collection of the information resulting from the PySkinDose analysis

    Attributes
    __________

    PSD : float
        The peak skin dose found in the dose map
    AirKerma : float
        The total air KERMA of the
    Events : EventOutput
        The event data for the examination
    Patient : dict[str, str | HumanPhantomOutput | NonHumanPhantomOutput]
        Information on the patient used in the calculations
    PatientOffsets : dict[str, float]
        The base offsets in the long, vert, and lat direction
    Table : Phantom
        The treatment table as an instance of the Phantom class
    Pad : Phantom
        The treatment table pad as an instance of the Phantom class
    PadThickness : float
        The thickness of the treatment table pad
    DoseMap : np.array
        The total dose map given as a numpy array where the values correspond to the resulting dose in Gy
    Hits : list[int, int]
        A list of all the hits of the radiation fields on the phantom cells given as a list of tuples on the form
        [(event_index, phantom_cell_index), ...]
    BackscatterCorrection : list[float]
        The backscatter corrections used for each cell hit given as a list of floats where the event and cell index of
        each float is given by getting the same list index element from the Hist attribute.
    InverseSquareLawCorrection : list[float]
        The inverse square law corrections used for each cell hit given as a list of floats where the event and cell
        index of each float is given by getting the same list index element from the Hist attribute.
    MediumCorrection : list[float]
        The corrections for the irradiated medium/-s used for each cell hit given as a list of floats where the event
        and cell index of each float is given by getting the same list index element from the Hist attribute.
    TableCorrection : list[float]
        The corrections for the treatment table used for each cell hit given as a list of floats where the event and
        cell index of each float is given by getting the same list index element from the Hist attribute.

    """

    def __init__(
        self,
        patient: Phantom,
        table: Phantom,
        pad: Phantom,
        dose_map: np.array,
        hits: list[list[float]],
        backscatter_correction: list[list[float]],
        inverse_square_law_correction: list[list[float]],
        medium_correction: list[float],
        table_correction: list[float],
        settings: PyskindoseSettings,
        data_norm: pd.DataFrame,
    ):
        """Create a PySkinDose output instance based on data from the PySkinDose run

        Parameters
        ----------
        patient : Phantom
            An instance of the Phantom class that represents the patient
        table : Phantom
            An instance of the Phantom class that represents the treatment table
        pad : Phantom
            An instance of the Phantom class that represents the pad
        dose_map : np.array
            A numpy array containing the calculated dose map
        hits : list[np.array]
            The numpy arrays containing information on which of the phantom cells are hit by the beam at each
            irradiation event
        backscatter_correction : list[list[float]]
            A list with a numpy array for each irradiation event containing the backscatter correction determined for
            each phantom cell
        inverse_square_law_correction : list[np.array]
            A list with a numpy array for each irradiation event containing the inverse square law correction determined
            for each phantom cell
        medium_correction : list[np.array]
            A list with a numpy array for each irradiation event containing the medium correction determined for
            each phantom cell
        table_correction : list[np.array]
            A list with a numpy array for each irradiation event containing the table correction determined for
            each phantom cell
        settings : PyskindoseSettings
            The instance of the settings class used in the PySkinDose run for the current data
        data_norm : pd.DataFrame
            The RDSR data, normalized for compliance with PySkinDose's use of units etc.
        """
        error = False
        error_message = [""]

        if len(backscatter_correction) != len(hits):
            error = True
            error_message.append(
                (
                    "Backscatter correction:\n"
                    "\tThe backscatter correction list is not the same length as the number of events"
                )
            )

        if len(inverse_square_law_correction) != len(hits):
            error = True
            error_message.append(
                (
                    "Inverse square law correction:\n"
                    "\tThe inverse square law correction list is not the same length as the number of events"
                )
            )

        if len(medium_correction) != len(hits):
            error = True
            error_message.append(
                ("Medium correction:\n" "\tThe medium correction list is not the same length as the number of events")
            )

        if len(table_correction) != len(hits):
            error = True
            error_message.append(
                ("Table correction:\n" "\tThe table correction list is not the same length as the number of events")
            )

        if error:
            raise ValueError("\n\n".join(error_message))

        self.PSD: float = dose_map.max()
        self.AirKerma: float = data_norm[KEY_NORMALIZATION_AIR_KERMA].sum()
        self.Events: EventOutput = EventOutput(data_norm=data_norm)
        self.PatientOffsets: dict = {
            "long": settings.phantom.patient_offset.d_lon,
            "vert": settings.phantom.patient_offset.d_ver,
            "lat": settings.phantom.patient_offset.d_lat,
        }
        self.Patient: dict = {
            "patient_type": patient.phantom_model,
            "patient": (
                HumanPhantomOutput(patient)
                if patient.phantom_model == PHANTOM_MODEL_HUMAN
                else NonHumanPhantomOutput(patient)
            ),
            "orientation": settings.phantom.patient_orientation,
        }
        self.Table: Phantom = table
        self.Pad: Phantom = pad
        self.PadThickness: float = settings.phantom.dimension.pad_thickness
        self.DoseMap: np.array = dose_map
        self.Hits = [[ind for ind, hit in enumerate(event_hits) if hit] for event_hits in hits]
        self.BackscatterCorrection = backscatter_correction
        self.InverseSquareLawCorrection = inverse_square_law_correction
        self.MediumCorrection = medium_correction
        self.TableCorrection = table_correction

    def to_dict(self) -> dict[str, Union[float, list[Union[float, int]]]]:
        """Converts the output data into a dict

        Returns
        -------
        Dict[str, Any]
            A dict containing the output data for the PySkinDose analysis where the lists of hits and corrections have
            been made sparse in order to save space.
        """
        return {
            "psd": self.PSD,
            "air_kerma": self.AirKerma,
            "patient": {
                "patient_type": self.Patient["patient_type"],
                "patient": self.Patient["patient"].to_dict(),
                "orientation": self.Patient["orientation"],
                "offsets": self.PatientOffsets,
            },
            "table": {
                "table_surface": {
                    "x": self.Table.r[:, 0].tolist(),
                    "y": self.Table.r[:, 1].tolist(),
                    "z": self.Table.r[:, 2].tolist(),
                },
                "triangle_vertex_indices": {
                    "i": self.Table.ijk[:, 0].tolist(),
                    "j": self.Table.ijk[:, 1].tolist(),
                    "k": self.Table.ijk[:, 2].tolist(),
                },
                "table_length": self.Table.table_length,
            },
            "pad": {
                "pad_surface": {
                    "x": self.Pad.r[:, 0].tolist(),
                    "y": self.Pad.r[:, 1].tolist(),
                    "z": self.Pad.r[:, 2].tolist(),
                },
                "triangle_vertex_indices": {
                    "i": self.Pad.ijk[:, 0].tolist(),
                    "j": self.Pad.ijk[:, 1].tolist(),
                    "k": self.Pad.ijk[:, 2].tolist(),
                },
            },
            "dose_map": [(ind, dose) for ind, dose in enumerate(self.DoseMap.tolist()) if dose > 0.0],
            "corrections": {
                "correction_value_index": self.Hits,
                "backscatter": self.BackscatterCorrection,
                "medium": self.MediumCorrection,
                "table": self.TableCorrection,
                "inverse_square_law": self.InverseSquareLawCorrection,
            },
            "events": self.Events.to_dict(),
        }

    def to_json(self) -> str:
        """Converts the output data into a JSON string

        Returns
        -------
        str
            A JSON formatted string containing the output data
        """
        return json.dumps(self.to_dict())


def format_analysis_result_for_export(
    analysis_result: Dict[str, Any],
    data_norm: pd.DataFrame,
    patient: Phantom,
    table: Phantom,
    pad: Phantom,
    settings: PyskindoseSettings,
) -> Union[PySkinDoseOutput, dict[str, Any], str]:
    """Formats the result of the PySkinDose analysis into a PySkinDoseOutput class instance that has a methods for
    converting the result to either a dict or a JSON string to facilitate building custom visualizations and for other
    custom implementations of the PySkinDose calculated data.

    Parameters
    ----------
    analysis_result : dict[str, Any]
        The dict resulting from the call to pyskindose.calculate_dose.calculate_dose.calculate_dose
    data_norm : pd.DataFrame
        The RDSR data, normalized for compliance with PySkinDose's use of units etc.
    patient : Phantom
        An instance of the Phantom class that represents the patient
    table : Phantom
        An instance of the Phantom class that represents the treatment table
    pad : Phantom
        An instance of the Phantom class that represents the pad
    settings : PyskindoseSettings
        The instance of the settings class used in the PySkinDose run for the current data

    Returns
    -------
    Union[PySkinDoseOutput, dict[str, Any], str]
        The PySkinDose formatted output as either a PySkinDoseOutput class instance, a dict or a JSON-formatted string
        depending on the output format specified in the settings
    """
    pyskindose_output = PySkinDoseOutput(
        patient=patient,
        table=table,
        pad=pad,
        dose_map=analysis_result[OUTPUT_KEY_DOSE_MAP],
        hits=[event if isinstance(event, list) else event.tolist() for event in analysis_result[OUTPUT_KEY_HITS]],
        backscatter_correction=[
            event if isinstance(event, list) else event.tolist()
            for event in analysis_result[OUTPUT_KEY_CORRECTION_BACK_SCATTER]
        ],
        inverse_square_law_correction=[
            event if isinstance(event, (list, float)) else ([] if event is None else event.tolist())
            for event in analysis_result[OUTPUT_KEY_CORRECTION_INVERSE_SQUARE_LAW]
        ],
        medium_correction=analysis_result[OUTPUT_KEY_CORRECTION_MEDIUM],
        table_correction=analysis_result[OUTPUT_KEY_CORRECTION_TABLE],
        settings=settings,
        data_norm=data_norm,
    )

    if settings.output_format == RUN_ARGUMENTS_OUTPUT_DICT:
        return pyskindose_output.to_dict()

    if settings.output_format == RUN_ARGUMENTS_OUTPUT_JSON:
        return pyskindose_output.to_json()

    return pyskindose_output



def _create_plotly_ijk_indices_for_cuboid_objects():

    # create list for storing indices order
    i_tab = []
    j_tab = []
    k_tab = []

    # append triangles to fill cuboid top side
    i_tab.append(0)
    j_tab.append(1)
    k_tab.append(2)

    i_tab.append(2)
    j_tab.append(3)
    k_tab.append(0)

    # append triangles to fill cuboid bottom side
    i_tab.append(0 + 4)
    j_tab.append(1 + 4)
    k_tab.append(2 + 4)

    i_tab.append(2 + 4)
    j_tab.append(3 + 4)
    k_tab.append(0 + 4)

    # append triangles to fill cuboid right side
    i_tab.append(0)
    j_tab.append(1)
    k_tab.append(4)

    i_tab.append(1)
    j_tab.append(5)
    k_tab.append(4)

    # append triangles to fill cuboid left side
    i_tab.append(2)
    j_tab.append(3)
    k_tab.append(6)

    i_tab.append(3)
    j_tab.append(7)
    k_tab.append(6)

    # append triangles to fill cuboid head side
    i_tab.append(0)
    j_tab.append(3)
    k_tab.append(4)

    i_tab.append(3)
    j_tab.append(7)
    k_tab.append(4)

    # append triangles to fill cuboid feet side
    i_tab.append(1)
    j_tab.append(2)
    k_tab.append(5)

    i_tab.append(2)
    j_tab.append(6)
    k_tab.append(5)

    return i_tab, j_tab, k_tab



import logging
from typing import List, Optional, Tuple

import numpy as np

from ..beam_class import Beam
from ..phantom_class import Phantom

logger = logging.getLogger(__name__)


def create_geometry_plot_texts(
    beam: Beam,
    table: Phantom,
    pad: Phantom,
    patient: Optional[Phantom] = None,
) -> Tuple[List[str], List[str], List[str], List[str], List[str], Optional[List[str]]]:
    """Create lists text strings to show when hovering in geometry plots

    Parameters
    ----------
    beam : Beam
        Instance of the Beam class containing info needed for the X-ray source, beam and
        detector texts
    table : Phantom
        Table phantom from instance of class Phantom. phantom_model must be "table"
    pad : Phantom
        Pad phantom from instance of class Phantom. phantom_model must be "pad"
    patient : Optional[Phantom]
        Patient phantom from instance of class Phantom. Can be of phantom_model "plane",
        "cylinder" or "human"
    """
    logger.debug("Creating geometry plot (and hover) texts")
    source_text = [
        (
            "<b>X-ray source</b>"
            "<br><br>"
            f"<b>LAT : </b>{np.around(beam.r[0, 2])} cm"
            "<br>"
            f"<b>LON : </b>{np.around(beam.r[0, 0])} cm"
            "<br>"
            f"<b>VER : </b>{np.around(beam.r[0, 1])} cm"
        )
    ]

    beam_text = [
        (
            "<b>X-ray beam vertex</b>"
            "<br><br>"
            f"<b>LAT : </b>{np.around(beam.r[ind, 2])} cm"
            "<br>"
            f"<b>LON : </b>{np.around(beam.r[ind, 0])} cm"
            "<br>"
            f"<b>VER : </b>{np.around(beam.r[ind, 1])} cm"
        )
        for ind in range(len(beam.r))
    ]

    detectors_text = [
        (
            "<b>X-ray detector</b>"
            "<br><br>"
            f"<b>LAT : </b>{np.around(beam.det_r[ind, 2])} cm"
            "<br>"
            f"<b>LON : </b>{np.around(beam.det_r[ind, 0])} cm"
            "<br>"
            f"<b>VER : </b>{np.around(beam.det_r[ind, 1])} cm"
        )
        for ind in range(len(beam.det_r))
    ]

    table_text = [
        (
            "<b>Support table</b>"
            "<br><br>"
            f"<b>LAT : </b>{np.around(table.r[ind, 2])} cm"
            "<br>"
            f"<b>LON : </b>{np.around(table.r[ind, 0])} cm"
            "<br>"
            f"<b>VER : </b>{np.around(table.r[ind, 1])} cm"
        )
        for ind in range(len(table.r))
    ]

    pad_text = [
        (
            "<b>Support pad</b>"
            f"<br><br><b>LAT : </b>{np.around(pad.r[ind, 2])} cm"
            "<br>"
            f"<b>LON : </b>{np.around(pad.r[ind, 0])} cm"
            "<br>"
            f"<b>VER : </b>{np.around(pad.r[ind, 1])} cm"
        )
        for ind in range(len(pad.r))
    ]

    patient_text = None
    if patient is not None:
        patient_text = [
            (
                "<b>Patient phantom</b>"
                f"<br><br><b>LAT : </b>{np.around(patient.r[ind, 2])} cm"
                "<br>"
                f"<b>LON : </b>{np.around(patient.r[ind, 0])} cm"
                "<br>"
                f"<b>VER : </b>{np.around(patient.r[ind, 1])} cm"
            )
            for ind in range(len(patient.r))
        ]

    return source_text, beam_text, detectors_text, table_text, pad_text, patient_text



import logging
from typing import List

import numpy as np
import pandas as pd

from pyskindose import Beam, Phantom
from pyskindose import constants as c
from pyskindose import scale_field_area
from pyskindose.corrections import calculate_k_isq
from pyskindose.geom_calc import check_table_hits

logger = logging.getLogger(__name__)


def perform_calculations_for_new_geometries(
    normalized_data: pd.DataFrame,
    event: int,
    new_geometry: bool,
    patient: Phantom,
    table: Phantom,
    pad: Phantom,
    hits: List[bool],
    table_hits: List[bool],
    field_area: List[float],
    k_isq: np.array,
):
    if not new_geometry:
        return hits, table_hits, field_area, k_isq

    beam = Beam(data_norm=normalized_data, event=event, plot_setup=False)

    patient.position(data_norm=normalized_data, event=event)
    table.position(data_norm=normalized_data, event=event)
    pad.position(data_norm=normalized_data, event=event)

    logger.debug("Checking which skin cells are hit by the beam")
    hits = beam.check_hit(patient=patient)

    if sum(hits):
        logger.debug("Checking which hit skin cells need table correction")
        table_hits = check_table_hits(source=beam.r[0, :], table=table, beam=beam, cells=patient.r[hits])

        logger.debug("Calculating X-Ray field area at the location of each skin cell")
        field_area = scale_field_area(
            data_norm=normalized_data,
            event=event,
            patient=patient,
            hits=hits,
            source=beam.r[0, :],
        )

        logger.debug("Calculating inverse-square law fluence correction")
        k_isq = calculate_k_isq(
            source=beam.r[0, :],
            cells=patient.r[hits],
            dref=normalized_data[c.DATA_DS_IRP][0],
        )

    return hits, table_hits, field_area, k_isq



def create_attributes_string(attrs_parent, object_name, indent_level, indent_size=4, indent_sign=" "):

    attrs_dict = vars(attrs_parent)

    indent_marker_title = (indent_level) * indent_size * indent_sign
    indent_marker_objs = indent_marker_title + indent_size * indent_sign

    attrs_str = indent_marker_title + object_name + "\n"

    for key, val in attrs_dict.items():
        if type(val) in [str, float, bool, int]:
            if not key == "attrs_str":
                attrs_str = attrs_str + indent_marker_objs + str(key) + " : " + str(val) + "\n"
            else:
                try:
                    attrs_parent_str = getattr(attrs_parent, key).attrs_str
                except AttributeError:
                    attrs_parent_str = ""
                attrs_str = f"{attrs_str}\n{attrs_parent_str}"

    return attrs_str



from pyskindose.helpers.create_attributes_string import create_attributes_string


class Plotsettings:
    """A class for setting plot settings.

    Attributes
    ----------
    interactivity : bool
        Toggle for interactive mode when plotting dosemaps. If True,
        the dosemap will be plotted in a .html file with full interactivity.
        If False, the dosemap will be saved as static images. Static mode is
        provided to enable PySkinDose to run smooth on machines with limited
        RAM.
    dark_mode : bool
        dark mode boolean
    notebook_mode : bool
        Set true if main is called from within a notebook.
        This optimizes plot sizing for notebook output cells.
    plot_dosemap : bool, default is True
        Whether dosemap should be plotted after dose calculation
    max_events_for_patient_inclusion : int
        maximum number of irradiation event for patient inclusion in
        plot_procedure. If the SR file contains more events than this number,
        the patient phantom is not shown in plot_procedure to avoid memory
        error.
    plot_event_index : int
        Index for the event that should be plotted when mode="plot_event" is
        chosen.

    """

    def __init__(self, plt_dict):
        """Initialize plot settings class.

        Parameters
        ----------
        plt_dict : dict
            Dictionary containing all of the plot setting that are
            appended as attributes to this class, see class attributes.
        """
        self.interactivity = plt_dict.get("interactivity", True)
        self.dark_mode = plt_dict.get("dark_mode", True)
        self.notebook_mode = plt_dict.get("notebook_mode", False)
        self.plot_dosemap = plt_dict.get("plot_dosemap", True)
        self.max_events_for_patient_inclusion = plt_dict.get("max_events_for_patient_inclusion", 10)
        self.plot_event_index = plt_dict.get("plot_event_index", 0)

        self.attrs_str = create_attributes_string(attrs_parent=self, object_name="plot", indent_level=0)

    def update_attrs_str(self):
        self.attrs_str = create_attributes_string(attrs_parent=self, object_name="plot", indent_level=0)

    def to_printable_string(self, color: str = "blue"):
        return (
            f"[bold {color}]Plot settings[/bold {color}]\n"
            f"\t[{color}]interactivity: {'True' if self.interactivity else 'False'}[/{color}]\n"
            f"\t[{color}]dark_mode: {'True' if self.dark_mode else 'False'}[/{color}]\n"
            f"\t[{color}]notebook_mode: {'True' if self.notebook_mode else 'False'}[/{color}]\n"
            f"\t[{color}]plot_dosemap: {'True' if self.plot_dosemap else 'False'}[/{color}]\n"
            f"\t[{color}]max_events_for_patient_inclusion: {self.max_events_for_patient_inclusion}[/{color}]\n"
            f"\t[{color}]plot_event_index: {self.plot_event_index}[/{color}]\n"
        )



from typing import Dict, Optional


class TranslationDirection:
    """Switch pos/neg direction of table translations.

    Attributes
    ----------
    keys of __init__ parameter directions
        Value of each attribute (i.e. key) is integers of either +1 or -1 to be
        used as a multiplicative correction factor to switch pos/neg direction
        in each direction. PySkinDose default angles are all +1.

    """

    x: Optional[int]
    y: Optional[int]
    z: Optional[int]

    def __init__(self, directions: Optional[Dict[str, str]] = None):
        """Initialize class attributes.

        Parameters
        ----------
        directions : dict
            dictionary with keys 'x', 'y' and 'z'.
            Each key contains either '+' or '-'.

        """
        self.x = None if directions is None else self._get_direction_as_value(directions["x"])
        self.y = None if directions is None else self._get_direction_as_value(directions["y"])
        self.z = None if directions is None else self._get_direction_as_value(directions["z"])

    def update_translation_direction(self, directions: Dict[str, str]):
        self.x = self._get_direction_as_value(directions["x"]) if self.x is None else self.x
        self.y = self._get_direction_as_value(directions["y"]) if self.y is None else self.y
        self.z = self._get_direction_as_value(directions["z"]) if self.z is None else self.z

    @staticmethod
    def _get_direction_as_value(direction: str):
        if direction not in ("+", "-"):
            raise ValueError(f"The direction must be given as '+' or '-' but was given as {direction}")

        return 1 if direction == "+" else -1



from typing import List

import numpy as np
import pandas as pd

from .phantom_class import Phantom


class Beam:
    """A class used to create an X-ray beam and detector.

    Attributes
    ----------
    r : np.array
        5*3 array, locates the xyz coordinates of the apex and verticies of a
        pyramid shaped X-ray beam, where the apex represents the X-ray focus
        (row 1) and the vertices where the beam intercepts the X-ray detector
        (row 2-5)

    ijk : np.array
        A matrix containing vertex indices. This is required in order to
        plot the beam using plotly Mesh3D. For more info, see "i", "j", and "k"
        at https://plot.ly/python/reference/#mesh3d
    det_r: np.array
        8*3 array, where each row locates the xyz coordinate of one of the 8
        corners of the cuboid shaped X-ray detector
    det_ijk : np.array
        same as ijk, but for plotting the X-ray detector
    N : np.array
        4*3 array, where each row contains a normal vector to one of the four
        faces of the beam.
    """

    def __init__(self, data_norm: pd.DataFrame, event: int = 0, plot_setup: bool = False) -> None:
        """Initialize the beam and detector for a specific irradiation event.

        Parameters
        ----------
        data_norm : pd.DataFrame
            Dicom RDSR information from each irradiation event. See
            rdsr_normalizer.py for more information.
        event : int, optional
            Specifies the index of the irradiation event in the procedure
            (the default is 0, which is the first event).
        plot_setup : bool, optional
            If True, the beam angulation info from data_norm is neglected,
            and a beam of zero angulation is created insted. This is a
            debugging feature used when positioning new phantoms or
            implementing currently unsupported venor RDSR files (the default is
            False).

        """
        # Override beam angulation if plot_setup
        if plot_setup:
            ap1 = ap2 = ap3 = 0

        else:
            # Fetch rotation angles of the X-ray tube

            # Positioner isocenter primary angle (Ap1)
            # i.e. rotation of the X-ray beam and detector about the z
            # axis (LAT)
            ap1 = np.deg2rad(data_norm.Ap1[event])
            # Positioner isocenter secondary angle (Ap2)
            # i.e. rotation of the X-ray beam and detector about the x axis
            # (LON)
            ap2 = np.deg2rad(data_norm.Ap2[event])
            # Positioner isocenter detector rotation angle (Ap3)
            # i.e. rotation of the X-ray detector about the y axis (VERT)
            ap3 = np.deg2rad(data_norm.Ap3[event])

        # calculate rotation about x axis
        angle = ap2
        Rx = np.array(
            [
                [+1, +0, +0],
                [+0, +np.cos(angle), -np.sin(angle)],
                [+0, +np.sin(angle), +np.cos(angle)],
            ]
        )

        # calculate rotation about y axis
        angle = ap3
        Ry = np.array(
            [
                [+np.cos(angle), +0, +np.sin(angle)],
                [+0, +1, +0],
                [-np.sin(angle), +0, +np.cos(angle)],
            ]
        )

        # calculate rotation about z axis
        angle = ap1
        Rz = np.array(
            [
                [+np.cos(angle), -np.sin(angle), +0],
                [+np.sin(angle), +np.cos(angle), +0],
                [+0, +0, +1],
            ]
        )

        # calculate source-isocenter displacement at ap1 = ap2 = 0
        delta_r = np.array([0, data_norm.DSI[event], 0])

        # Create unit-beam in the positioner coordinate system
        r = np.array(
            [
                [0, 0, 0],  # r0 (i.e. X-ray source)
                [+0.5, -1.0, +0.5],  # r+
                [+0.5, -1.0, -0.5],  # r+-
                [-0.5, -1.0, -0.5],  # r-
                [-0.5, -1.0, +0.5],  # r-+
            ]
        )
        r[1:, 1] *= data_norm.DSD[event]
        r[1:, 0] *= data_norm.FS_long[event]  # Append longitudinal collimation
        r[1:, 2] *= data_norm.FS_lat[event]  # Append lateral collimation

        # Transform the beam from the positioner coordinate system to the
        # isocenter coordinate system. Note! The transpose operations are
        # needed to broadcast over all vectors in r
        r = np.matmul(Rz, np.matmul(Rx, (r + delta_r).T)).T

        self.r = r

        # Manually create vertex index vector for the X-ray beam
        self.ijk = np.column_stack(([0, 0, 0, 0, 1, 1], [1, 1, 3, 3, 2, 3], [2, 4, 2, 4, 3, 4]))

        # Create unit vectors from X-ray source to beam verticies
        v = ((self.r[1:] - self.r[0, :]).T / np.linalg.norm(self.r[1:] - self.r[0, :], axis=1)).T

        # Create the four normal vectors to the faces of the beam.
        self.N = np.vstack(
            [
                np.cross(v[0, :], v[1, :]),
                np.cross(v[1, :], v[2, :]),
                np.cross(v[2, :], v[3, :]),
                np.cross(v[3, :], v[0, :]),
            ]
        )

        # Create detector corners for with side length 1
        # The first four rows represent the X-ray detector surface, the last
        # four are there to give the detector some depth for 3D visualization.
        det_r = np.array(
            [
                [+0.5, -1.0, +0.5],
                [+0.5, -1.0, -0.5],
                [-0.5, -1.0, -0.5],
                [-0.5, -1.0, +0.5],
                [+0.5, -1.2, +0.5],
                [+0.5, -1.2, -0.5],
                [-0.5, -1.2, -0.5],
                [-0.5, -1.2, +0.5],
            ]
        )

        # Add detector dimensions
        detector_width = data_norm.DSL[0]
        det_r[:, 0] *= detector_width
        det_r[:, 2] *= detector_width
        # Place detector at actual distance
        det_r[:, 1] *= data_norm.DID[event]

        # Transform the detector from the positioner coordinate system to
        # the isocenter coordinate system. Note! The transpose operations
        # are needed to broadcast over all vector in r_det
        det_r = np.matmul(Rz, np.matmul(Ry, np.matmul(Rx, det_r.T))).T

        self.det_r = det_r

        # Manually construct vertex index vector for the X-ray detector
        self.det_ijk = np.column_stack(
            (
                [0, 0, 4, 4, 0, 1, 0, 3, 3, 7, 1, 1],
                [1, 2, 5, 6, 1, 5, 3, 7, 2, 2, 2, 6],
                [2, 3, 6, 7, 4, 4, 4, 4, 7, 6, 6, 5],
            )
        )

    def check_hit(self, patient: Phantom) -> List[bool]:
        """Calculate which patient entrance skin cells are hit by the beam.

        A description of this algoritm is presented in the wiki, please visit
        https://pyskindose.readthedocs.io/en/latest/

        Parameters
        ----------
        patient : Phantom
            Patient phantom, either of type plane, cylinder or human, i.e.
            instance of class Phantom

        Returns
        -------
        List[bool]
            A boolean list of the same length as the number of patient skin
            cells. True for all entrance skin cells that are hit by the beam.

        """
        # Create vectors from X-ray source to each phantom skin cell
        v = patient.r - self.r[0, :]

        # Check which skin cells lies within the beam
        hits = (np.dot(v, self.N.T) <= 0).all(axis=1)
        # if patient phantom is 3D, remove exit path skin cells
        if patient.phantom_model != "plane":
            temp1 = v[hits]
            temp2 = patient.n[hits]

            bool_entrance = [np.dot(temp1[i], temp2[i]) <= 0 for i in range(len(temp1))]

            hits[np.where(hits)] = bool_entrance

        return hits.tolist()



from pyskindose.constants import (
    OFFSET_LATERAL_KEY,
    OFFSET_LONGITUDINAL_KEY,
    OFFSET_VERTICAL_KEY,
)
from pyskindose.helpers.create_attributes_string import create_attributes_string


class PatientOffset:
    """A class for setting patient - table offset.

    In PyskinDose, the table isocenter is located centered at the head end
    of the support table. The attributes in this class is used to offset the
    patient phantom from this isocenter, in order to get correct patient
    positioning.

    Attributes
    ----------
    d_lat : int
        latertal offset from table isocenter
    d_ver : int
        Vertical offset from table isocenter
    d_lon : int
        longitudianl offset from table isocenter

    Raises
    ------
    NotImplementedError
            Raises error if other units then cm are used.

    """

    def __init__(self, offset: dict):
        """Initialize patient-table offset class.

        Parameters
        ----------
        offset : dict
            offset in cm from the table isocenter in the lateral, vertical and
            longitudinal direction.

        """
        self.d_lat = offset[OFFSET_LATERAL_KEY]
        self.d_ver = offset[OFFSET_VERTICAL_KEY]
        self.d_lon = offset[OFFSET_LONGITUDINAL_KEY]

        self.attrs_str = create_attributes_string(attrs_parent=self, object_name="patient offset", indent_level=1)

    def update_attrs_str(self):
        self.attrs_str = create_attributes_string(attrs_parent=self, object_name="patient offset", indent_level=1)



import logging
import os
from pathlib import Path

import pandas as pd
import pydicom

from pyskindose import rdsr_normalizer, rdsr_parser
from pyskindose.settings import PyskindoseSettings

logger = logging.getLogger(__name__)


def read_and_normalise_rdsr_data(rdsr_filepath: str, settings: PyskindoseSettings):

    rdsr_filepath = (
        Path(rdsr_filepath)
        if rdsr_filepath
        else Path(__file__).parent.parent / "example_data/RDSR" / settings.rdsr_filename
    )

    logger.debug(str(rdsr_filepath.absolute()))

    # If provided, load preparsed rdsr data in .json format
    if ".json" == rdsr_filepath.suffix.lower():
        return pd.read_json(rdsr_filepath)

    # else load RDSR data with pydicom
    data_raw = pydicom.dcmread(rdsr_filepath)

    # parse RDSR data from raw .dicom file
    data_parsed = rdsr_parser(data_raw, silence_pydicom_warnings=settings.silence_pydicom_warnings)

    # normalized rdsr for compliance with PySkinDose
    normalized_data = rdsr_normalizer(data_parsed, settings=settings)

    if settings.remove_invalid_rows:
        if invalid_kvp_rows := len(normalized_data[normalized_data.kVp == 0]):
            print(f"Removing {invalid_kvp_rows} rows with kVp value = 0")
            normalized_data = normalized_data[normalized_data.kVp != 0].reset_index(drop=True)

    return normalized_data



import logging
from pathlib import Path
from typing import List

import plotly.graph_objects as go
from PIL import Image

from ..constants import (
    PLOT_BASE_SIZE_STATIC,
    PLOT_GROUND_SHIFT_X_STATIC,
    PLOT_SHIFT_X_STATIC,
)

logger = logging.getLogger(__name__)


def create_notebook_dose_map_plot(names: List[Path]):

    fig = go.Figure()

    # Constants
    ground_shift_x = PLOT_GROUND_SHIFT_X_STATIC
    shift_x = PLOT_SHIFT_X_STATIC
    base_size_static = PLOT_BASE_SIZE_STATIC

    img_width = base_size_static + ground_shift_x / 4
    img_height = base_size_static + ground_shift_x / 4
    nrows = 1
    ncols = 4

    total_width = ncols * img_width - ground_shift_x
    total_height = nrows * img_height

    fig.add_trace(go.Scatter(x=[0, total_width], y=[0, img_height], mode="lines", visible=False))

    placements = [0, 1, 2, 3]
    placements = [img_width * place for place in placements]
    placements = [place - ground_shift_x for place in placements]
    placements = [placements[i] - shift_x[i] for i in range(len(shift_x))]

    for i, placement in enumerate(placements):

        source = Image.open(names[i].absolute())

        # Add image
        fig.add_layout_image(
            dict(
                x=placement,
                sizex=img_width,
                y=1 * img_height,
                sizey=img_height,
                xref="x",
                yref="y",
                opacity=1,
                layer="above",
                sizing="stretch",
                source=source,
            )
        )

    axes_visability = False

    # Configure axes
    fig.update_xaxes(
        visible=axes_visability,
        range=[0, total_width],
    )

    fig.update_yaxes(visible=axes_visability, range=[0, total_height], scaleanchor="x")

    fig.update_layout(
        width=total_width,
        height=total_height,
        margin={"l": 0, "r": 0, "t": 0, "b": 0},
    )

    fig.show(config={"doubleClick": "reset"})

    return



from pathlib import Path
from typing import List

import pandas as pd
import spekpy as sp  # Note: Not included in PySkinDose,
from tqdm import tqdm


def generate_hvl_data(
    kvp_range: List[float],
    filtration_inherent_mmal: List[float],
    filtration_added_mmal: List[float],
    filtration_added_mmcu: List[float],
    anode_angle_deg: int,
    path_results: Path,
    filter_name: str,
):
    """Generate HVL data with SpekPy for k_bs and k_med selection.

    Parameters
    ----------
    kvp_range : List[float]
        list with range of kvp vals (in kV)
    filtration_inherent_mmal : float
        list of range of X-ray tube inherent filtration (in mmAl)
    filtration_added_mmal : List[float]
        X-ray tube added Al filtration. E.g. [0] for AXIOM-Artis
        or [0, 1] for Allura Clarity
    filtration_added_mmcu : List[float]
        X-ray tube added Cu filtration for spekpy spectrum generation. E.g. [0, 0.1, 0.2, 0.3, 0.6, 0.9] for AXIOM-Artis
        or [0, 0.1, 0.4, 0.9] for Allura Clarity
    anode_angle_deg : int
        Anode angle for spekpy spectrum generation. E.g. 8 for "AXIOM-Artis" or 11 for "Allura Clarity"
    path_results : Path
        The path to store results
    filter_name : str
        tag in filename for filter alternative, e.g. "allura" or "artis".

    """
    index = [
        "kvp_kv",
        "filtration_inherent_mmal",
        "filtration_added_mmal",
        "filtration_added_mmcu",
        "anode_angle_deg",
        "hvl_mmal",
    ]

    res = dict()
    for i in index:
        res[i] = []

    for kvp in tqdm(kvp_range):
        for inherent_mmal in filtration_inherent_mmal:
            for added_mmal in filtration_added_mmal:
                for added_mmcu in filtration_added_mmcu:
                    s = sp.Spek(kvp=kvp, th=anode_angle_deg, dk=1)
                    filters = [("Al", inherent_mmal + added_mmal), ("Cu", added_mmcu)]
                    s.multi_filter(filters)

                    res["kvp_kv"].append(kvp)
                    res["filtration_inherent_mmal"].append(inherent_mmal)
                    res["filtration_added_mmal"].append(added_mmal)
                    res["filtration_added_mmcu"].append(added_mmcu)
                    res["anode_angle_deg"].append(anode_angle_deg)
                    res["hvl_mmal"].append(s.get_hvl())

    file_name = f"hvl_{filter_name}_{anode_angle_deg}deg.csv"
    file_name = file_name.replace(" ", "")

    pd.DataFrame(res).to_csv(path_results / file_name, index=False)



import pandas as pd

from pyskindose.phantom_class import Phantom

from .plot_event import plot_event
from .plot_procedure import plot_procedure
from .plot_setup import plot_setup


def plot_geometry(
    patient: Phantom,
    table: Phantom,
    pad: Phantom,
    data_norm: pd.DataFrame,
    mode: str,
    event: int = 0,
    dark_mode: bool = True,
    notebook_mode: bool = False,
    include_patient: bool = False,
) -> None:
    """Visualize the geometry from the irradiation events.

    This function plots the phantom, table, pad, beam and detector. The type of plot is
    specified in the parameter mode.

    Parameters
    ----------
    patient : Phantom
        Patient phantom from instance of class Phantom. Can be of phantom_model "plane",
        "cylinder" or "human"
    table : Phantom
        Table phantom from instance of class Phantom. phantom_model must be "table"
    pad : Phantom
        Pad phantom from instance of class Phantom. phantom_model must be "pad"
    data_norm : pd.DataFrame
        Table containing dicom RDSR information from each irradiation event. See
        rdsr_normalizer.py for more information.
    mode : str
        Choose between "plot_setup", "plot_event" and "plot_procedure".

         "plot_setup" plots the patient, table and pad in reference position together
         with the X-ray system with zero angulation. This is a debugging feature

        "plot_event" plots a specific irradiation event, in which the patient, table,
        pad, and beam are positioned according to the irradiation event specified by the
        parameter event

        "plot_procedure" plots out the X-ray system, table and pad for all irradiation
        events in the procedure The visible event are chosen by a event slider

    event : int, optional
        choose specific irradiation event if mode "plot_event" are used. Default is 0,
        in which the first irradiation event is considered.

    dark_mode : bool, optional
        set dark mode for plots

    notebook_mode : bool, optional
        optimize plot size for notebooks, default is True.

    include_patient : bool, optional
        Choose if the patient phantom should be included in the plot_procedure function.
        WARNING, very heavy on memory. Default is False.

    """
    plot_setup(
        mode=mode,
        data_norm=data_norm,
        patient=patient,
        table=table,
        pad=pad,
        dark_mode=dark_mode,
        notebook_mode=notebook_mode,
    )

    plot_event(
        mode=mode,
        data_norm=data_norm,
        event=event,
        patient=patient,
        table=table,
        pad=pad,
        dark_mode=dark_mode,
        notebook_mode=notebook_mode,
    )

    plot_procedure(
        mode=mode,
        data_norm=data_norm,
        include_patient=include_patient,
        patient=patient,
        table=table,
        pad=pad,
        dark_mode=dark_mode,
        notebook_mode=notebook_mode,
    )



import json
from pathlib import Path
from typing import Optional, Union

from rich import print

from pyskindose.constants import (
    KEY_PARAM_ESTIMATE_K_TAB,
    KEY_PARAM_INHERENT_FILTRATION,
    KEY_PARAM_K_TAB_VAL,
    KEY_PARAM_MODE,
    KEY_PARAM_RDSR_FILENAME,
    KEY_PARAM_REMOVE_INVALID_ROWS,
    KEY_PARAM_SILENCE_PYDICOM_WARNINGS,
    RUN_ARGUMENTS_OUTPUT_DICT,
    RUN_ARGUMENTS_OUTPUT_HTML,
    RUN_ARGUMENTS_OUTPUT_JSON,
    RUN_ARGUMENTS_VALID_OUTPUT_FORMATS,
)

from .normalization_settings import NormalizationSettings
from .phantom_settings import PhantomSettings
from .plot_settings import Plotsettings


class PyskindoseSettings:
    """A class to store all settings required to run PySkinDose.

    Attributes
    ----------
    mode : str
        Select which mode to execute PySkinDose with. There are three
        different modes:

        mode = "calculate_dose" calculates the skin dose from the RDSR data and
        presents the result in a skin dose map.

        mode = "plot_setup" plots the geometry (patient, table, pad and beam
        in starting position, i.e., before any RDSR data has been added.) This
        is useful for debugging and when manually fixating the patient phantom
        with the function "position_patient_phantom_on_table".

        mode = "plot_event" plots the geometry for a specific irradiation event
        with index = event.

        mode = "plot_procedure" plots geometry of the entire sequence of RDSR
        events provided in the RDSR file. The patient phantom is omitted for
        calculation speed in human phantom is used.

    rdsr_filename : str
        filename of the RDSR file, without the .dcm file ending.
    estimate_k_tab : bool
        Whether k_tab should be approximated or not. You should set this to true if you
        have not conducted table attenuation measurements.
    k_tab_val : float
        Value of k_tab, in range 0.0 -> 1.0.
    inherent_filtration : float
        X-ray tube inherent filtration, for backscatter and medium correction.
    phantom : pyskindose.settings.phantom_settings.PhantomSettings
        Instance of class PhantomSettings containing all phantom related
        settings.
    plot : pyskindose.settings.plot_settings.Plotsettings
        Instance of class Plotsettings containing all plot related settings

    """

    def __init__(
        self,
        settings: Union[str, dict],
        normalization_settings: Optional[Union[Path, str, dict, NormalizationSettings]] = None,
        file_result_output_path: Optional[Union[str, Path]] = None,
        output_format: str = RUN_ARGUMENTS_OUTPUT_HTML,
    ):
        """Initialize settings class.

        Parameters
        ----------
        settings : Union[str, dict]
            Either a JSON-string or a dictionary containing all the settings
            parameters required to run PySkinDose. See setting_example.json
            in /settings/ for example.

        """
        if isinstance(settings, str):
            tmp = json.loads(settings)
        else:
            tmp = settings

        if (output_format := output_format.lower()) not in RUN_ARGUMENTS_VALID_OUTPUT_FORMATS:
            raise ValueError(
                f"The output format must be specified as one of {', '.join(RUN_ARGUMENTS_VALID_OUTPUT_FORMATS)}"
            )

        self.mode = tmp[KEY_PARAM_MODE]
        self.output_format = output_format
        self.file_result_output_path: Path = self._initialize_output_path(
            output_path=file_result_output_path, output_format=output_format
        )
        self.k_tab_val = tmp[KEY_PARAM_K_TAB_VAL]
        self.inherent_filtration = tmp[KEY_PARAM_INHERENT_FILTRATION]
        self.silence_pydicom_warnings = tmp[KEY_PARAM_SILENCE_PYDICOM_WARNINGS]
        self.rdsr_filename = tmp[KEY_PARAM_RDSR_FILENAME]
        self.estimate_k_tab = tmp[KEY_PARAM_ESTIMATE_K_TAB]
        self.phantom = PhantomSettings(ptm_dim=tmp["phantom"])
        self.plot = Plotsettings(plt_dict=tmp["plot"])
        self.corrections_db_path = tmp.get("corrections_db_path", "corrections.db")

        self.normalization_settings = self._initialize_normalization_settings(normalization_settings)

        self.remove_invalid_rows: bool = True if tmp.get(KEY_PARAM_REMOVE_INVALID_ROWS) else None

    @staticmethod
    def _initialize_output_path(output_path: Optional[Union[str, Path]], output_format: str) -> Path:
        if output_path is None:
            output = Path.cwd() / "PlotOutputs"

            if output_format in (RUN_ARGUMENTS_OUTPUT_DICT, RUN_ARGUMENTS_OUTPUT_JSON):
                return output  # Return without creating the output directory as it won't be used

            output.mkdir(exist_ok=True)
            return output

        if isinstance(output_path, str):
            output_path = Path(output_path)

        if not isinstance(output_path, Path):
            raise TypeError("file_result_output_path must be a string or a Path object")

        if output_path.is_dir():
            return output_path

        raise ValueError("file_result_output_path must be a path to a directory")

    @staticmethod
    def _initialize_normalization_settings(
        normalization_settings: Optional[Union[Path, str, dict, NormalizationSettings]]
    ) -> NormalizationSettings:
        if normalization_settings is None:
            normalization_settings = Path(__file__).parent.parent / "normalization_settings.json"

        if isinstance(normalization_settings, Path):
            normalization_settings = normalization_settings.read_text()

        if isinstance(normalization_settings, str):
            normalization_settings = json.loads(normalization_settings)

        if isinstance(normalization_settings, dict):
            if "normalization_settings" in normalization_settings:
                normalization_settings = normalization_settings["normalization_settings"]
            normalization_settings = NormalizationSettings(normalization_settings)

        if isinstance(normalization_settings, NormalizationSettings):
            return normalization_settings

        raise TypeError(f"Invalid type {type(normalization_settings)} given for normalization_settings")

    def print_parameters(self, return_as_string: bool = False):
        """Print entire parameter class to terminal.

        Parameters
        ----------
        return_as_string : bool, optional
            Return the print statement as a string, instead of printing it
            to the terminal. The default is False.

        """
        phantom_settings_string = self.phantom.to_printable_string(color="bright_magenta")
        plot_settings_string = self.plot.to_printable_string(color="steel_blue1")
        normalization_settings_string = self.normalization_settings.to_printable_string(color="bright_green")

        color = "bright_cyan"

        output_str = (
            f"[b u {color}]General settings[/b u {color}]\n"
            f"\t[{color}]mode:\t{self.mode}[/{color}]\n"
            f"\t[{color}]rdsr_filename:\t{self.rdsr_filename}[/{color}]\n"
            f"\t[{color}]estimate_k_tab:\t{'True' if self.estimate_k_tab else 'False'}[/{color}]\n"
            f"\t[{color}]silence_pydicom_warnings:\t{'True' if self.silence_pydicom_warnings else 'False'}[/{color}]\n"
            f"\n{phantom_settings_string}"
            f"\n{plot_settings_string}"
            f"\n{normalization_settings_string}"
        )

        if return_as_string:
            return output_str

        return print(output_str)


def initialize_settings(settings: Union[str, dict, PyskindoseSettings]) -> PyskindoseSettings:
    valid_input_settings = settings is not None and isinstance(settings, (str, dict, PyskindoseSettings))

    if not valid_input_settings:
        raise ValueError("Settings must be given as a str or dict")

    if isinstance(settings, PyskindoseSettings):
        return settings

    return PyskindoseSettings(settings=settings)



import logging
from pathlib import Path
from typing import Optional

from pyskindose.settings import PyskindoseSettings, initialize_settings

logger = logging.getLogger(__name__)


def parse_settings_to_settings_class(settings: Optional[str] = None):
    try:
        return initialize_settings(settings)
    except ValueError:
        logger.debug("Tried initializing settings without any settings")

    settings_path = Path(__file__).parent.parent / "settings.json"

    if not settings_path.exists():
        logger.warning("Settings path not specified. Using example settings.")
        settings_path = settings_path.parent / "settings_example.json"

    return PyskindoseSettings(settings_path.read_text())



import pandas as pd
import pydicom

from pyskindose.constants import (
    KEY_RDSR_ACQUISITION_DATA,
    KEY_RDSR_COMMENT,
    KEY_RDSR_CONCEPT_CODE_SEQUENCE,
    KEY_RDSR_CONTENT_SEQUENCE,
    KEY_RDSR_DETECTORSIZE_MM,
    KEY_RDSR_EVENT_XRAY_DATA,
    KEY_RDSR_II_DIAMETER_SRDATA,
    KEY_RDSR_MANUFACTURER,
    KEY_RDSR_MANUFACTURER_MODEL_NAME,
    KEY_RDSR_MEASURED_VALUE_SEQUENCE,
    KEY_RDSR_TEXT_VALUE,
    KEY_RDSR_UID,
)


def rdsr_parser(data_raw: pydicom.FileDataset, silence_pydicom_warnings=False) -> pd.DataFrame:
    """Parse event data from radiation dose structure reports (RDSR).

    Parameters
    ----------
    data_raw : pydicom.FileDataset
        RDSR file from fluoroscopic device, opened with package pydicom.

    Returns
    -------
    pd.DataFrame
        Parsed RDSR data from all irradiation events in the RDSR input file

    """
    if silence_pydicom_warnings:
        pydicom.config.settings.reading_validation_mode = pydicom.config.IGNORE

    # create list to store rdsr content from each irradiation event
    prodcedure_dicts = []
    # For each content in RDSR file
    for rdsr_content in data_raw.ContentSequence:

        # If content = irradiation event
        if rdsr_content.ConceptNameCodeSequence[0].CodeMeaning == KEY_RDSR_EVENT_XRAY_DATA:

            # Declare temporary dictionary
            data_parsed_dict = dict()

            # Save manufacturer, and manufacturer model name
            data_parsed_dict[KEY_RDSR_MANUFACTURER] = data_raw.Manufacturer
            data_parsed_dict[KEY_RDSR_MANUFACTURER_MODEL_NAME] = data_raw.ManufacturerModelName

            # For each content in 'Irradiation Event X-Ray Data'
            for xray_event_content in rdsr_content.ContentSequence:
                # Reformat 'Concept Name'
                tag = (
                    xray_event_content.ConceptNameCodeSequence[0]
                    .CodeMeaning.replace(" ", "")
                    .replace("-", "")
                    .replace("(", "")
                    .replace(")", "")
                    .replace(".", "")
                )

                # Save 'Concept Name' to dictionary, assign corresponding value
                if KEY_RDSR_CONCEPT_CODE_SEQUENCE in xray_event_content:
                    if tag in data_parsed_dict.keys():
                        data_parsed_dict[tag] = [
                            data_parsed_dict[tag],
                            xray_event_content.ConceptCodeSequence[0].CodeMeaning,
                        ]
                    else:
                        data_parsed_dict[tag] = xray_event_content.ConceptCodeSequence[0].CodeMeaning

                elif KEY_RDSR_MEASURED_VALUE_SEQUENCE in xray_event_content:
                    # If the content contains a 'Measured Value Sequence'
                    # Reformat 'Concept Name' to include unit of measurement
                    unit = (
                        xray_event_content.MeasuredValueSequence[0]
                        .MeasurementUnitsCodeSequence[0]
                        .CodeValue.replace(".", "")
                    )

                    tag = "_".join([tag, unit])

                    if tag in data_parsed_dict.keys():
                        data_parsed_dict[tag] = [
                            data_parsed_dict[tag],
                            xray_event_content.MeasuredValueSequence[0].NumericValue,
                        ]

                    else:
                        data_parsed_dict[tag] = xray_event_content.MeasuredValueSequence[0].NumericValue

                elif KEY_RDSR_TEXT_VALUE in xray_event_content:

                    # This loop extracts detector size for static acquisitions,
                    # which is given as a 'Comment' for siemens artis zee units
                    if tag == KEY_RDSR_COMMENT:
                        comment = xray_event_content.TextValue.split("/")
                        if KEY_RDSR_ACQUISITION_DATA in comment[0]:
                            for index in comment:
                                if KEY_RDSR_II_DIAMETER_SRDATA in index:
                                    data_parsed_dict[KEY_RDSR_DETECTORSIZE_MM] = index.split("=")[1].replace('"', "")

                    elif tag in data_parsed_dict.keys():
                        data_parsed_dict[tag] = [data_parsed_dict[tag], xray_event_content.TextValue]
                    else:
                        data_parsed_dict[tag] = xray_event_content.TextValue

                elif KEY_RDSR_UID in xray_event_content:
                    if tag in data_parsed_dict.keys():
                        data_parsed_dict[tag] = [data_parsed_dict[tag], xray_event_content.UID]
                    else:
                        data_parsed_dict[tag] = xray_event_content.UID

                # If the 'Irradiation Event X-Ray Data' contains subcontent
                elif KEY_RDSR_CONTENT_SEQUENCE in xray_event_content:
                    # For each subcontent
                    for xray_event_subcontent in xray_event_content.ContentSequence:
                        # Reformat 'Concept Name'
                        tag = (
                            xray_event_subcontent.ConceptNameCodeSequence[0]
                            .CodeMeaning.replace(" ", "")
                            .replace("-", "")
                            .replace("(", "")
                            .replace(")", "")
                            .replace(".", "")
                        )

                        # corresponding value
                        if KEY_RDSR_CONCEPT_CODE_SEQUENCE in xray_event_subcontent:
                            if tag in data_parsed_dict.keys():
                                data_parsed_dict[tag] = [
                                    data_parsed_dict[tag],
                                    xray_event_subcontent.ConceptCodeSequence[0].CodeMeaning,
                                ]
                            else:
                                data_parsed_dict[tag] = xray_event_subcontent.ConceptCodeSequence[0].CodeMeaning

                        elif KEY_RDSR_TEXT_VALUE in xray_event_subcontent:

                            if tag in data_parsed_dict.keys():
                                data_parsed_dict[tag] = [data_parsed_dict[tag], xray_event_subcontent.TextValue]

                            else:
                                data_parsed_dict[tag] = xray_event_subcontent.TextValue

                        elif KEY_RDSR_UID in xray_event_subcontent:
                            if tag in data_parsed_dict.keys():
                                data_parsed_dict[tag] = [data_parsed_dict[tag], xray_event_subcontent.UID]
                            else:
                                data_parsed_dict[tag] = xray_event_subcontent.UID
                        elif KEY_RDSR_MEASURED_VALUE_SEQUENCE in xray_event_subcontent:
                            # Reformat 'Concept Name' to include unit of
                            # measurement
                            unit = (
                                xray_event_subcontent.MeasuredValueSequence[0].MeasurementUnitsCodeSequence[0].CodeValue
                            )

                            tag = "_".join([tag, unit])

                            if tag in data_parsed_dict.keys():
                                data_parsed_dict[tag] = (
                                    data_parsed_dict[tag],
                                    xray_event_subcontent.MeasuredValueSequence[0].NumericValue,
                                )
                            else:
                                data_parsed_dict[tag] = xray_event_subcontent.MeasuredValueSequence[0].NumericValue

                        # Assign None to 'Concept Name' if nothing relevant to
                        # parse in RDSR subcontent
                        else:
                            data_parsed_dict[tag] = None

                # Assign None to 'Concept Name' if nothing relevant to parse
                # in RDSR content
                else:
                    data_parsed_dict[tag] = None

            # Store event info
            prodcedure_dicts.append(data_parsed_dict)

    data_parsed = pd.DataFrame(prodcedure_dicts)

    return data_parsed



import logging

import numpy as np
import pandas as pd

from pyskindose.settings.normalization_settings import NormalizationSettings

from .constants import (
    KEY_NORMALIZATION_ACQUISITION_PLANE,
    KEY_NORMALIZATION_ACQUISITION_TYPE,
    KEY_NORMALIZATION_AIR_KERMA,
    KEY_NORMALIZATION_DISTANCE_ISOCENTER_DETECTOR,
    KEY_NORMALIZATION_DISTANCE_SOURCE_DETECTOR,
    KEY_NORMALIZATION_DISTANCE_SOURCE_IRP,
    KEY_NORMALIZATION_DISTANCE_SOURCE_ISOCENTER,
    KEY_NORMALIZATION_FILTER_SIZE_ALUMINUM,
    KEY_NORMALIZATION_FILTER_SIZE_COPPER,
    KEY_NORMALIZATION_MODEL_NAME,
    KEY_RDSR_DISTANCE_SOURCE_DETECTOR,
    KEY_RDSR_FILTER_MATERIAL,
    KEY_RDSR_FILTER_MATERIAL_ALUMINUM,
    KEY_RDSR_FILTER_MATERIAL_COPPER,
    KEY_RDSR_FILTER_MAX,
    KEY_RDSR_FILTER_MIN,
)
from .geom_calc import calculate_field_size
from .settings import PyskindoseSettings

logger = logging.getLogger("pyskindose")


def rdsr_normalizer(data_parsed: pd.DataFrame, settings: PyskindoseSettings) -> pd.DataFrame:
    """Normalize RDSR data for PySkinDose compliance.

    Parameters
    ----------
    data_parsed : pd.DataFrame
        Parsed RDSR data from all irradiation events in the RDSR input file,
        i.e. output of function rdsr_parser
    settings : PyskindoseSettings
        The PySkinDose settings object containing the normalization settings

    Returns
    -------
    data_norm : pd.DataFrame
        DataFrame with the following columns
            - model (str)
                device model, e.g. 'AXIOM-artis
            - DSD (float)
                Distance Source to detector (DSD) in cm.
            - DSI (float)
                Distance Source to Isocenter (DSI) in cm.
            - DID (float)
                Distance Isocenter to Detector (DID) in cm.
            - DSIRP (float)
                Distance Source to intercentional reference point (DSIRP) in
                cm.
            - acquisition_type (str)
                Type of irradiation event, i.e. 'fluoroscopy, or stationary
                acquisition.
            - acquisition_plane (str)
                plane used for image acquisition. Either 'single plane',
                'plane a', or 'plane b'.
            - Tx (float)
                Table offset in x-direction (longitudinal direction) from the
                machine isocenter. At Tx = 0, the patient support table is
                centered about the isocenter x-axis. With patient lying in
                head-first supine position (default settings), Tx increases in
                patient left lateral direction.
            - Ty (float)
                Table offset in y-direction (vertical direction) from the
                machine isocenter. At Ty = 0, the patient support table is
                entered about the isocenter y-axis. Ty is increasing downwards,
                i.e. along the force of gravity.
            - Tz (float)
                Table offset in z-direction (lateral direction) from the
                machine isocenter. At Tz = 0, the head end of the patient
                support tables are located at the zero coordinate of the
                z-axis. With patient lying in head-first supine position
                (default settings), Tz increases in the patients cranial
                direction.

                .. image:: user/figures/table/table_translate.svg

            - At1 (int)
                Rotation angle of the patient support table about the isocenter
                y-axis. The center of rotation is located at the centerpoint
                of the table. Positive direction is determined by
                the right-hand rule for curve orientation about the positive
                isocenter y-axis.

                .. image:: user/figures/table/table_at1.svg

            -At2 (int)
                Tilt angel of the patient support table about the isocenter
                x-axis. The center of the tilt is located at the center of the
                table, with positive direction determined by the right-hand
                rule for curve orientations  about the positiove isocenter
                x-axis.

                .. image:: user/figures/table/table_at2.svg

            - At3 (int)
                Cradle angle of the patient support table about the isocenter
                z-axis. The center of rotation is located at the centerpoint
                of the table. Positive direction is determined by the
                right-hand rule for curve orientation about the positive
                isocenter z-axis.

                .. image:: user/figures/table/table_at3.svg

            - Ap1 (int)
                Rotation angle of the X-ray source about the isocenter z-axis.
                Positive direction is determined by the right-hand rule for
                curve orientation about the positive isocenter z-axis.

                .. image:: user/figures/beam/beam_ap1.svg

            - Ap2 (int)
                Rotation angle of the X-ray source about the isocenter x-axis.
                Positive direction is determined by the right-hand rule for
                curve orientation about the positive isocenter x-axis.

                .. image:: user/figures/beam/beam_ap2.svg

    Ap3 : int
        Rotation angle of the X-ray detector about the isocenter y-axis.
        Positive direction is determined by the right-hand rule for curve
        orientation about the positive isocenter y-axis.


    DSL : float
        Detector Side Length (DSL) in cm.
    FS_lat : float
        Side length of the X-ray field in the lateral direction at the image
        receptor plane.
    FS_long : float
        Side length of the X-ray field in the longitudinal direction at the
        image receptor plane.
    kVp : float
        Tube voltage in kV
    K_IRP : float
        IRP air kerma at the Interventional Reference Point (IRP).
    filter_thickness_Cu : float
        Copper X-ray filter thickness in mm.
    filter_thickness_Al : float
        Aluminum X-ray filter thickness in mm.
    """
    data_norm = pd.DataFrame()

    settings.normalization_settings.update_used_settings(data_parsed=data_parsed)

    for append_normalization in [
        _normalize_machine_parameters,
        _normalize_table_parameters,
        _normalize_xray_filter_materials,
        _normalize_beam_parameters,
    ]:
        data_norm = append_normalization(
            data_parsed=data_parsed, data_norm=data_norm, norm=settings.normalization_settings
        )

    return data_norm


def _normalize_machine_parameters(
    data_parsed: pd.DataFrame, data_norm: pd.DataFrame, norm: NormalizationSettings
) -> pd.DataFrame:

    data_norm[KEY_NORMALIZATION_MODEL_NAME] = data_parsed.ManufacturerModelName

    # Find indices of nans in DistanceSourcetoDetector
    if "nan" in str(data_parsed[KEY_RDSR_DISTANCE_SOURCE_DETECTOR]).lower():
        nan_indices = data_parsed.index[data_parsed[KEY_RDSR_DISTANCE_SOURCE_DETECTOR].apply(np.isnan)]
        # Replace those nans with the corresponding value in
        # FinalDistanceSourcetoDetector
        data_parsed.DistanceSourcetoDetector_mm = data_parsed.DistanceSourcetoDetector_mm.fillna(
            data_parsed.FinalDistanceSourcetoDetector_mm[nan_indices]
        )

    data_norm[KEY_NORMALIZATION_DISTANCE_SOURCE_DETECTOR] = data_parsed.DistanceSourcetoDetector_mm / 10
    data_norm[KEY_NORMALIZATION_DISTANCE_SOURCE_ISOCENTER] = data_parsed.DistanceSourcetoIsocenter_mm / 10
    data_norm[KEY_NORMALIZATION_DISTANCE_ISOCENTER_DETECTOR] = data_norm.DSD - data_norm.DSI
    data_norm[KEY_NORMALIZATION_DISTANCE_SOURCE_IRP] = data_norm.DSI - 15
    data_norm[KEY_NORMALIZATION_ACQUISITION_TYPE] = data_parsed.IrradiationEventType
    data_norm[KEY_NORMALIZATION_ACQUISITION_PLANE] = data_parsed.AcquisitionPlane

    return data_norm


def _normalize_table_parameters(
    data_parsed: pd.DataFrame, data_norm: pd.DataFrame, norm: NormalizationSettings
) -> pd.DataFrame:

    # Table translations
    data_norm["Tx"] = norm.trans_offset.x + norm.trans_dir.x * data_parsed.TableLongitudinalPosition_mm / 10

    data_norm["Ty"] = norm.trans_offset.y + norm.trans_dir.y * data_parsed.TableHeightPosition_mm / 10

    data_norm["Tz"] = norm.trans_offset.z + norm.trans_dir.z * data_parsed.TableLateralPosition_mm / 10

    # Table rotations
    data_norm["At1"] = norm.rot_dir.At1 * [0] * len(data_norm)
    data_norm["At2"] = norm.rot_dir.At2 * [0] * len(data_norm)
    data_norm["At3"] = norm.rot_dir.At3 * [0] * len(data_norm)

    return data_norm


def _normalize_xray_filter_materials(
    data_parsed: pd.DataFrame, data_norm: pd.DataFrame, norm: NormalizationSettings
) -> pd.DataFrame:
    # parse filter material and thickness

    # Load filter min and max, and fill all NANs with zeros
    for key in [KEY_RDSR_FILTER_MIN, KEY_RDSR_FILTER_MAX]:
        data_parsed[key] = data_parsed[key].fillna(0.0)

    # Add columns for filter materials in data_norm, and initialize to zero
    for key in [
        KEY_NORMALIZATION_FILTER_SIZE_COPPER,
        KEY_NORMALIZATION_FILTER_SIZE_ALUMINUM,
    ]:
        data_norm[key] = 0.0

    # for each irradiation event
    for event_index in range(len(data_parsed)):
        # fetch filter materials from data_parsed
        event_filter_materials = data_parsed[KEY_RDSR_FILTER_MATERIAL][event_index]

        # fetch filter min and max thicknesses
        event_filter_minmax = np.array(
            [
                data_parsed[KEY_RDSR_FILTER_MIN][event_index],
                data_parsed[KEY_RDSR_FILTER_MAX][event_index],
            ]
        )

        # calculate filter mean thicknesses
        event_filter_means = np.mean(event_filter_minmax, axis=0)

        if isinstance(event_filter_materials, str):
            event_filter_means = [event_filter_means]

        if not isinstance(event_filter_means, (list, np.ndarray)) and np.isnan(event_filter_materials):
            logger.debug("Skipping mean value filter thickness calculation for event with no filter")
            continue

        # if copper filter in use
        if KEY_RDSR_FILTER_MATERIAL_COPPER in event_filter_materials:
            # append copper filtration to normalized data
            data_norm.loc[event_index, KEY_NORMALIZATION_FILTER_SIZE_COPPER] = event_filter_means[
                event_filter_materials.index(KEY_RDSR_FILTER_MATERIAL_COPPER)
            ]
        # if aluminum filter in use
        if KEY_RDSR_FILTER_MATERIAL_ALUMINUM in event_filter_materials:
            # append aluminum filtration to normalized data
            data_norm.loc[event_index, KEY_NORMALIZATION_FILTER_SIZE_ALUMINUM] = event_filter_means[
                event_filter_materials.index(KEY_RDSR_FILTER_MATERIAL_ALUMINUM)
            ]

    return data_norm


def _normalize_beam_parameters(
    data_parsed: pd.DataFrame, data_norm: pd.DataFrame, norm: NormalizationSettings
) -> pd.DataFrame:

    # beam angulation
    data_norm["Ap1"] = norm.rot_dir.Ap1 * data_parsed.PositionerPrimaryAngle_deg
    data_norm["Ap2"] = norm.rot_dir.Ap2 * data_parsed.PositionerSecondaryAngle_deg
    # temp set to zero
    data_norm["Ap3"] = norm.rot_dir.Ap3 * [0] * len(data_norm)

    # detector side length
    data_norm["DSL"] = norm.detector_side_length

    FS_lat, FS_long = calculate_field_size(
        field_size_mode=norm.field_size_mode,
        data_parsed=data_parsed,
        data_norm=data_norm,
    )

    data_norm["FS_lat"] = FS_lat
    data_norm["FS_long"] = FS_long

    data_norm["kVp"] = data_parsed.KVP_kV
    data_norm[KEY_NORMALIZATION_AIR_KERMA] = data_parsed.DoseRP_Gy * 1000

    return data_norm



import logging
from typing import Any, Dict, List, Optional

import pandas as pd
import plotly.graph_objects as go

from ..constants import (
    COLOR_SLIDER_BACKGROUND,
    IRRADIATION_EVENT_STEP_KEY_ARGUMENTS,
    IRRADIATION_EVENT_STEP_KEY_LABEL,
    IRRADIATION_EVENT_STEP_KEY_METHOD,
    MODE_PLOT_PROCEDURE,
    PLOT_ASPECTMODE_PLOT_PROCEDURE,
    PLOT_AXIS_TITLE_X,
    PLOT_AXIS_TITLE_Y,
    PLOT_AXIS_TITLE_Z,
    PLOT_DRAGMODE,
    PLOT_FONT_FAMILY,
    PLOT_FONT_SIZE,
    PLOT_HOVERLABEL_FONT_FAMILY,
    PLOT_PROCEDURE_AXIS_RANGE_X,
    PLOT_PROCEDURE_AXIS_RANGE_Y,
    PLOT_PROCEDURE_AXIS_RANGE_Z,
    PLOT_SLIDER_BORDER_WIDTH,
    PLOT_SLIDER_FONT_SIZE_CURRENT,
    PLOT_SLIDER_FONT_SIZE_GENERAL,
    PLOT_SLIDER_TRANSITION,
    PLOT_TITLE_FONT_FAMILY,
    PLOT_TITLE_FONT_SIZE,
    PLOT_ZERO_LINE_WIDTH,
)
from ..phantom_class import Phantom
from .create_irradiation_event_procedure_plot_data import (
    create_irradiation_event_procedure_plot_data,
)
from .create_plot_and_save_to_file import create_plot_and_save_to_file
from .get_camera_view import get_camera_view
from .plot_settings import (
    fetch_plot_colors,
    fetch_plot_margin,
    fetch_plot_size,
    fetch_slider_colors,
    fetch_slider_padding,
)

logger = logging.getLogger(__name__)


def plot_procedure(
    mode: str,
    data_norm: pd.DataFrame,
    table: Phantom,
    pad: Phantom,
    include_patient: bool,
    patient: Optional[Phantom] = None,
    dark_mode: bool = True,
    notebook_mode: bool = False,
):
    """Create plot_procedure plot.

    Parameters
    ----------
    mode : str
        The function will only run if this is set to "plot_procedure".
    data_norm : pd.DataFrame
        RDSR data, normalized for compliance with PySkinDose.
    table : Phantom
        Patient support table phantom
    pad : Phantom
        Patient support pad phantom
    include_patient : bool
        Choose if the patient phantom should be included
    patient : Optional[Phantom], optional
        patient phantom, by default None
    dark_mode : bool, optional
        set dark mode for plots, by default True
    notebook_mode : bool, optional
        optimize plot size for notebooks, default is True.

    Raises
    ------
    IOError
        Raises error if patient not provided when include_patient = True

    """
    if mode != MODE_PLOT_PROCEDURE:
        return

    if include_patient and patient is None:
        logger.error("Plot procedure called with include patient but no patient input")
        raise IOError("Patient object must be given when include_patient set to True")

    logger.info(f"Plotting entire procedure with {len(data_norm)} irradiation events")

    title = f"<b>P</b>y<b>S</b>kin<b>D</b>ose [mode: {mode}]"

    meshes = [
        create_irradiation_event_procedure_plot_data(
            data_norm=data_norm,
            include_patient=include_patient,
            visible_status=(ind == 0),
            event=ind,
            patient=(patient if include_patient else None),
            table=table,
            pad=pad,
        )
        for ind in range(len(data_norm))
    ]

    data = [event.get(plot_object) for plot_object in meshes[0].keys() for event in meshes]

    layout = _create_procedure_layout(
        title=title,
        total_events=len(data_norm),
        dark_mode=dark_mode,
        notebook_mode=notebook_mode,
    )

    create_plot_and_save_to_file(mode=mode, data=data, layout=layout)


def _create_event_slider_step(total_events: int, event: int) -> Dict[str, Any]:

    step = {
        IRRADIATION_EVENT_STEP_KEY_METHOD: "restyle",
        IRRADIATION_EVENT_STEP_KEY_ARGUMENTS: ["visible", [False] * total_events],
        IRRADIATION_EVENT_STEP_KEY_LABEL: event + 1,
    }
    step[IRRADIATION_EVENT_STEP_KEY_ARGUMENTS][1][event] = True

    return step


def _create_sliders(
    steps: List[Dict],
    total_events: int,
    dark_mode: bool = True,
    notebook_mode: bool = False,
) -> List[Dict[str, Any]]:

    COLOR_PLOT_TEXT, COLOR_SLIDER_TICK, COLOR_SLIDER_BORDER = fetch_slider_colors(dark_mode=dark_mode)

    PLOT_SLIDER_PADDING = fetch_slider_padding(notebook_mode=notebook_mode)

    return [
        dict(
            active=0,
            transition=PLOT_SLIDER_TRANSITION,
            bordercolor=COLOR_SLIDER_BORDER,
            borderwidth=PLOT_SLIDER_BORDER_WIDTH,
            tickcolor=COLOR_SLIDER_TICK,
            bgcolor=COLOR_SLIDER_BACKGROUND,
            currentvalue=dict(
                prefix="Active event: ",
                suffix=f" of {total_events}",
                font=dict(color=COLOR_PLOT_TEXT, size=PLOT_SLIDER_FONT_SIZE_CURRENT),
            ),
            font=dict(
                family=PLOT_FONT_FAMILY,
                color=COLOR_PLOT_TEXT,
                size=PLOT_SLIDER_FONT_SIZE_GENERAL,
            ),
            pad=PLOT_SLIDER_PADDING,
            steps=steps,
        )
    ]


def _create_procedure_layout(
    title: str, total_events: int, dark_mode: bool = True, notebook_mode: bool = False
) -> go.Layout:

    steps = [_create_event_slider_step(total_events=total_events, event=ind) for ind in range(total_events)]

    COLOR_CANVAS, COLOR_PLOT_TEXT, COLOR_GRID, COLOR_ZERO_LINE = fetch_plot_colors(dark_mode=dark_mode)

    PLOT_HEIGHT, PLOT_WIDTH = fetch_plot_size(notebook_mode=notebook_mode)

    PLOT_MARGIN = fetch_plot_margin(notebook_mode=notebook_mode)

    layout = go.Layout(
        height=PLOT_HEIGHT,
        width=PLOT_WIDTH,
        margin=PLOT_MARGIN,
        sliders=_create_sliders(
            steps=steps,
            total_events=total_events,
            dark_mode=dark_mode,
            notebook_mode=notebook_mode,
        ),
        font=dict(family=PLOT_FONT_FAMILY, size=PLOT_FONT_SIZE, color=COLOR_PLOT_TEXT),
        hoverlabel=dict(font=dict(family=PLOT_HOVERLABEL_FONT_FAMILY, size=PLOT_SLIDER_FONT_SIZE_GENERAL)),
        showlegend=False,
        dragmode=PLOT_DRAGMODE,
        title=dict(
            text=title,
            font=dict(
                family=PLOT_TITLE_FONT_FAMILY,
                size=PLOT_TITLE_FONT_SIZE,
                color=COLOR_PLOT_TEXT,
            ),
        ),
        paper_bgcolor=COLOR_CANVAS,
        scene=dict(
            aspectmode=PLOT_ASPECTMODE_PLOT_PROCEDURE,
            camera=get_camera_view(),
            xaxis=dict(
                title=PLOT_AXIS_TITLE_X,
                range=PLOT_PROCEDURE_AXIS_RANGE_X,
                color=COLOR_PLOT_TEXT,
                gridcolor=COLOR_GRID,
                linecolor=COLOR_GRID,
                backgroundcolor=COLOR_CANVAS,
                zerolinecolor=COLOR_ZERO_LINE,
                zerolinewidth=PLOT_ZERO_LINE_WIDTH,
            ),
            yaxis=dict(
                title=PLOT_AXIS_TITLE_Y,
                range=PLOT_PROCEDURE_AXIS_RANGE_Y,
                color=COLOR_PLOT_TEXT,
                gridcolor=COLOR_GRID,
                linecolor=COLOR_GRID,
                backgroundcolor=COLOR_CANVAS,
                zerolinecolor=COLOR_ZERO_LINE,
                zerolinewidth=PLOT_ZERO_LINE_WIDTH,
            ),
            zaxis=dict(
                title=PLOT_AXIS_TITLE_Z,
                range=PLOT_PROCEDURE_AXIS_RANGE_Z,
                color=COLOR_PLOT_TEXT,
                gridcolor=COLOR_GRID,
                linecolor=COLOR_GRID,
                backgroundcolor=COLOR_CANVAS,
                zerolinecolor=COLOR_ZERO_LINE,
                zerolinewidth=PLOT_ZERO_LINE_WIDTH,
            ),
        ),
    )

    return layout



import logging
from typing import List, Union

import plotly.graph_objects as go

logger = logging.getLogger(__name__)


def create_plot_and_save_to_file(mode: str, data: List[Union[go.Mesh3d, go.Scatter3d]], layout: go.Layout):
    """

    :param mode:
    :param data:
    :param layout:
    :return:
    """
    plot_filename = f"{mode}.html"

    logger.debug(f"Creating plot and savint to file {plot_filename}")

    fig = go.Figure(data=data, layout=layout)

    fig.show()



import os
from pathlib import Path

import pandas as pd

file_names = [
    "hvl_tables/hvl_allura_filters_11deg.csv",
    "hvl_tables/hvl_axiom_filters_8deg.csv",
]

data_path = Path(os.path.abspath(__file__)).parent

data = [pd.read_csv(data_path / file_name) for file_name in file_names]
data_combined = pd.concat([dat for dat in data])

data_combined.to_csv(data_path / "hvl_tables/hvl_combined.csv", index=False)



import logging
from typing import Dict, Optional, Union

import pandas as pd
import plotly.graph_objects as go

from ..beam_class import Beam
from ..constants import (
    COLOR_BEAM,
    COLOR_DETECTOR,
    COLOR_PAD,
    COLOR_PATIENT,
    COLOR_SOURCE,
    COLOR_TABLE,
    IRRADIATION_EVENT_PROCEDURE_KEY_BEAM,
    IRRADIATION_EVENT_PROCEDURE_KEY_DETECTORS,
    IRRADIATION_EVENT_PROCEDURE_KEY_PAD,
    IRRADIATION_EVENT_PROCEDURE_KEY_PATIENT,
    IRRADIATION_EVENT_PROCEDURE_KEY_SOURCE,
    IRRADIATION_EVENT_PROCEDURE_KEY_TABLE,
    IRRADIATION_EVENT_PROCEDURE_KEY_WIRE_FRAME_BEAM,
    IRRADIATION_EVENT_PROCEDURE_KEY_WIRE_FRAME_DETECTORS,
    IRRADIATION_EVENT_PROCEDURE_KEY_WIRE_FRAME_PAD,
    IRRADIATION_EVENT_PROCEDURE_KEY_WIRE_FRAME_TABLE,
    MESH_OPACITY_BEAM,
)
from ..phantom_class import Phantom
from .create_geometry_plot_texts import create_geometry_plot_texts
from .create_mesh3d import create_mesh_3d_general
from .create_wireframes import create_wireframes

logger = logging.getLogger(__name__)


def create_irradiation_event_procedure_plot_data(
    data_norm: pd.DataFrame,
    include_patient: bool,
    visible_status: bool,
    event: int,
    table: Phantom,
    pad: Phantom,
    patient: Optional[Phantom] = None,
) -> Dict[str, Union[go.Scatter3d, go.Mesh3d]]:

    # Position geometry objects
    beam = Beam(data_norm, event=event, plot_setup=False)
    table.position(data_norm, event)
    pad.position(data_norm, event)

    (
        source_text,
        beam_text,
        detectors_text,
        table_text,
        pad_text,
        patient_text,
    ) = create_geometry_plot_texts(beam=beam, table=table, pad=pad, patient=patient)

    output = {}

    if include_patient:
        patient.position(data_norm, event)

        # Create patient mesh
        output[IRRADIATION_EVENT_PROCEDURE_KEY_PATIENT] = create_mesh_3d_general(
            obj=patient,
            color=COLOR_PATIENT,
            mesh_text=patient_text,
            lighting=dict(diffuse=0.5, ambient=0.5),
            visible_status=visible_status,
        )

    # Create X-ray source mesh
    output[IRRADIATION_EVENT_PROCEDURE_KEY_SOURCE] = go.Scatter3d(
        x=[beam.r[0, 0], beam.r[0, 0]],
        y=[beam.r[0, 1], beam.r[0, 1]],
        z=[beam.r[0, 2], beam.r[0, 2]],
        mode="markers",
        hoverinfo="text",
        visible=visible_status,
        marker=dict(size=8, color=COLOR_SOURCE),
        text=source_text,
    )

    # Create support table mesh
    output[IRRADIATION_EVENT_PROCEDURE_KEY_TABLE] = create_mesh_3d_general(
        obj=table,
        color=COLOR_TABLE,
        mesh_text=table_text,
        visible_status=visible_status,
    )

    # Create X-ray detector mesh
    output[IRRADIATION_EVENT_PROCEDURE_KEY_DETECTORS] = create_mesh_3d_general(
        obj=beam,
        color=COLOR_DETECTOR,
        mesh_text=detectors_text,
        visible_status=visible_status,
        detector_mesh=True,
    )

    # Create support pad mesh
    output[IRRADIATION_EVENT_PROCEDURE_KEY_PAD] = create_mesh_3d_general(
        obj=pad, color=COLOR_PAD, mesh_text=pad_text, visible_status=visible_status
    )

    # Create X-ray beam mesh
    output[IRRADIATION_EVENT_PROCEDURE_KEY_BEAM] = create_mesh_3d_general(
        obj=beam,
        color=COLOR_BEAM,
        mesh_text=beam_text,
        visible_status=visible_status,
        opacity=MESH_OPACITY_BEAM,
    )

    # Create wireframe meshes
    [
        output[IRRADIATION_EVENT_PROCEDURE_KEY_WIRE_FRAME_BEAM],
        output[IRRADIATION_EVENT_PROCEDURE_KEY_WIRE_FRAME_TABLE],
        output[IRRADIATION_EVENT_PROCEDURE_KEY_WIRE_FRAME_PAD],
        output[IRRADIATION_EVENT_PROCEDURE_KEY_WIRE_FRAME_DETECTORS],
    ] = create_wireframes(beam=beam, table=table, pad=pad, line_width=4, visible=visible_status)

    return output



from typing import Dict, Optional


class TranslationOffset:
    """Set translation offset of patient support table.

    Use this class to set the translation offset (in cm) between the machine
    origin (the unit that generated the RDSR) and the machine origin of
    PySkinDose (which is located at (x,y,z) = (0, 0, 0)).

    Attributes
    ----------
    keys of __init__ parameter offset
        Value of each attribute (i.e. key) is a float specifying the offset
        in cm.

    """

    def __init__(self, offset: Optional[Dict[str, float]] = None):
        """Initialize class attributes.

        Parameters
        ----------
        offset : dict
            dictionary with keys 'x', 'y' and 'z'. Each key contains the
            translation offset (in that direction), specified as a float in cm.

        """
        if offset is None:
            offset = dict()

        self.x: float = offset.get("x")
        self.y: float = offset.get("y")
        self.z: float = offset.get("z")

    def update_translation_offset(self, offset: Dict[str, float]):
        self.x = float(offset["x"]) if self.x is None else self.x
        self.y = float(offset["y"]) if self.y is None else self.y
        self.z = float(offset["z"]) if self.z is None else self.z



from typing import Dict, List, Optional, Union

import plotly.graph_objects as go

from ..beam_class import Beam
from ..phantom_class import Phantom
from .get_visual_offset import _get_visual_offset


def create_mesh_3d_general(
    obj: Union[Phantom, Beam],
    color: str,
    mesh_text: List[str],
    opacity: Optional[float] = None,
    mesh_name: Optional[str] = None,
    lighting: Optional[Dict] = None,
    detector_mesh: bool = False,
    visible_status: Optional[bool] = True,
) -> go.Mesh3d:

    if opacity is None:
        opacity = 1.0

    visual_offset = _get_visual_offset(patient=obj)

    mesh_x = obj.r[:, 0]
    mesh_y = obj.r[:, 1] + visual_offset
    mesh_z = obj.r[:, 2]
    mesh_i = obj.ijk[:, 0]
    mesh_j = obj.ijk[:, 1]
    mesh_k = obj.ijk[:, 2]

    if detector_mesh:
        mesh_x = obj.det_r[:, 0]
        mesh_y = obj.det_r[:, 1] + visual_offset
        mesh_z = obj.det_r[:, 2]
        mesh_i = obj.det_ijk[:, 0]
        mesh_j = obj.det_ijk[:, 1]
        mesh_k = obj.det_ijk[:, 2]

    if lighting is None and mesh_name is None:
        return go.Mesh3d(
            x=mesh_x,
            y=mesh_y,
            z=mesh_z,
            i=mesh_i,
            j=mesh_j,
            k=mesh_k,
            color=color,
            hoverinfo="text",
            text=mesh_text,
            opacity=opacity,
            visible=visible_status,
        )

    if mesh_name is None:
        return go.Mesh3d(
            x=mesh_x,
            y=mesh_y,
            z=mesh_z,
            i=mesh_i,
            j=mesh_j,
            k=mesh_k,
            color=color,
            hoverinfo="text",
            text=mesh_text,
            opacity=opacity,
            lighting=lighting,
            visible=visible_status,
        )

    if lighting is None:
        return go.Mesh3d(
            x=mesh_x,
            y=mesh_y,
            z=mesh_z,
            i=mesh_i,
            j=mesh_j,
            k=mesh_k,
            color=color,
            hoverinfo="text",
            text=mesh_text,
            name=mesh_name,
            opacity=opacity,
            visible=visible_status,
        )

    return go.Mesh3d(
        x=mesh_x,
        y=mesh_y,
        z=mesh_z,
        i=mesh_i,
        j=mesh_j,
        k=mesh_k,
        color=color,
        hoverinfo="text",
        text=mesh_text,
        name=mesh_name,
        lighting=lighting,
        opacity=opacity,
        visible=visible_status,
    )



import logging
from typing import List

import numpy as np
import pandas as pd
import scipy.interpolate
from scipy.interpolate import CubicSpline

from .db_connect import db_connect

logger = logging.getLogger(__name__)


def calculate_k_isq(source: np.array, cells: np.array, dref: float) -> np.array:
    """Calculate the IRP air kerma inverse-square law correction.

    This function corrects the X-ray fluence from the interventionl reference point
    (IRP), to the actual source to skin distance, so that the IRP air kerma is converted
    to air kerma at the patient skin surface.

    Parameters
    ----------
    source : np.array
        location of the X-ray source
    cells : np.array
        location of all the cells that are hit by the beam
    dref : float
        reference distance source to IRP, i.e. the distance at which the IRP air kerma
        is stated.

    Returns
    -------
    np.array
        Inverse-square law correction for all cells that are hit by the beam.

    """
    if len(cells) > 3:
        return np.square(dref / np.linalg.norm(cells - source, axis=1))

    return np.square(dref / np.linalg.norm(cells - source, axis=0))


def calculate_k_bs(data_norm: pd.DataFrame) -> List[CubicSpline]:
    """Calculate backscatter correction.

    This function calculates the backscatter correction factor for all events, at field
    sizes [5, 10, 20, 25, 35] cm^2. The function uses the non-linear interpolation
    method presented by
    Benmakhlouf et al. in the article "Influence of phantom thickness and material on
    the backscatter factors for diagnostic x-ray beam dosimetry",
    [doi:10.1088/0031-9155/58/2/247]

    Parameters
    ----------
    data_norm : pd.DataFrame
        RDSR data, normalized for compliance with PySkinDose.

    Returns
    -------
    List[CubicSpline]
        List of scipy cubic spline interpolation object for all events.

    """
    # Tabulated field side length in cm
    fsl_tab = [5, 10, 20, 25, 35]

    # polynomial coefficents
    c = np.array(
        [
            [+1.00870e0, +9.29969e-1, +8.65442e-1, +8.58665e-1, +8.57065e-1],
            [+2.35816e-3, +4.08549e-3, +5.36739e-3, +5.51579e-3, +5.55933e-3],
            [-9.48937e-6, -1.66271e-5, -2.21494e-5, -2.27532e-5, -2.28004e-5],
            [+1.03143e-1, +1.53605e-1, +1.72418e-1, +1.70826e-1, +1.66418e-1],
            [-1.04881e-3, -1.45187e-3, -1.46088e-3, -1.38540e-3, -1.28180e-3],
            [+3.59731e-6, +5.05312e-6, +5.17430e-6, +4.91192e-6, +4.53036e-6],
            [-7.31303e-3, -9.32427e-3, -8.30138e-3, -7.64330e-3, -6.81574e-3],
            [+7.93272e-5, +9.40568e-5, +7.13576e-5, +6.13126e-5, +4.94197e-5],
            [-2.74296e-7, -3.28449e-7, -2.54885e-7, -2.21399e-7, -1.79074e-7],
        ]
    )

    # Fetch kVp and HVL from data_norm
    kvp = data_norm.kVp
    hvl = data_norm.HVL

    # Calculate k_bs for field side length [5, 10, 20, 25, 35] cm
    # This is eq. (8) in doi:10.1088/0031-9155/58/2/247.
    bs_corr = [
        (c[0, :] + c[1, :] * kvp[event] + c[2, :] * np.square(kvp[event]))
        + (c[3, :] + c[4, :] * kvp[event] + c[5, :] * np.square(kvp[event])) * hvl[event]
        + (c[6, :] + c[7, :] * kvp[event] + c[8, :] * np.square(kvp[event])) * np.square(hvl[event])
        for event in range(len(kvp))
    ]

    # Create interpolation object for bs_corr
    bs_interp = [scipy.interpolate.CubicSpline(fsl_tab, bs_corr[event]) for event in range(len(kvp))]

    return bs_interp


def calculate_k_med(data_norm: pd.DataFrame, field_area: List[float], event: int, corrections_db: str) -> float:
    """Calculate medium correction.

    This function calculates and appends the medium correction factor for all skin cells
    that are hit by the X-ray beam in an event. The correction factors are from the
    article "Backscatter factors and mass energy-absorption coefficient ratios for
    surface dose determination in diagnostic radiology".

    Parameters
    ----------
    data_norm : pd.DataFrame
        RDSR data, normalized for compliance with PySkinDose.
    field_area : List[float]
        X-ray field area in (cm^2) for each phantom skin cell that are hit by the X-ray
        beam.
    event : int
        Irradiation event index.
    corrections_db : str
        A string defining the path to the corrections SQLite db


    Returns
    -------
    float
        Medium correction k_med for all cells that are hit by the beam.

    """
    # Tabulated field side length in cm
    fsl_tab = [5, 10, 20, 25, 35]

    # Fetch kVp and HVL from data_norm
    kvp = data_norm.kVp[event]
    hvl = data_norm.HVL[event]

    # Calculate mean side length for all cells that are hit by the beam.
    # This field size dependance of k_med is negligible (<= 1%), therefore,
    # independep field size resolution is omitted for computational speed.
    fsl_mean = np.mean(np.sqrt(field_area))

    # Select the closest available tabulated field size length.
    fsl = min(fsl_tab, key=lambda x: abs(x - fsl_mean))

    # Connect to database
    conn = db_connect(db_name=corrections_db)[0]

    # Fetch k_med = f(kVp, HVL) from database. This is table 2 in
    # [doi:10.1088/0031-9155/58/2/247]
    df = pd.read_sql_query(
        """SELECT kvp_kv, hvl_mmal, field_side_length_cm,
                           mu_en_quotient FROM correction_medium_and_backscatter""",
        conn,
    )

    conn.commit()
    conn.close()

    # Fetch kVp entries from table
    kvp_data = df.loc[(df["field_side_length_cm"] == fsl), "kvp_kv"]
    # Select closest tabulated kVp (strongest dependence for k_med)
    kvp_round = min(kvp_data, key=lambda x: abs(x - kvp))

    # Fetch HVL entries from table
    hvl_data = df.loc[(df["field_side_length_cm"] == fsl) & (df["kvp_kv"] == kvp_round), "hvl_mmal"]

    # Select closest tabulated HVL (second strongest dependence for k_med)
    hvl_round = min(hvl_data, key=lambda x: abs(x - hvl))

    # Fetch corresponding k_med
    k_med = float(
        df.loc[
            (df["hvl_mmal"] == hvl_round) & (df["kvp_kv"] == kvp_round) & (df["field_side_length_cm"] == fsl),
            "mu_en_quotient",
        ].iloc[0]
    )

    return k_med


def calculate_k_tab(
    data_norm: pd.DataFrame, corrections_db: str, estimate_k_tab: bool = False, k_tab_val: float = 0.8
) -> List[float]:
    """Fetch table correction factor from database.

    This function fetches measured table correction factor as a function of
    HVL and kVp. Further, if no measurement are conducted on a specific unit,
    the function can also return user specified estimated table correction.

    Parameters
    ----------
    data_norm : pd.DataFrame
        RDSR data, normalized for compliance with PySkinDose.
    estimate_k_tab: bool
        Set to True to use estimated table correction, default is False.
    k_tab_val: float
        Value of estimated table corrections, must be in range (0, 1).
    corrections_db : str
        A string defining the path to the corrections SQLite db

    Returns
    -------
    List[float]
        List of table correction factor for all events in procedure.

    """
    if estimate_k_tab:
        return [k_tab_val] * len(data_norm)

    # Connect to database
    [conn, c] = db_connect(db_name=corrections_db)

    k_tab = [1.0] * len(data_norm)

    try:
        # For every irradiation event
        for event in range(len(data_norm)):

            # Set parameters for fetching table transmission correction factor.
            params = (
                round(float(data_norm.kVp[event])),
                float(data_norm.filter_thickness_Cu[event]),
                float(data_norm.filter_thickness_Al[event]),
                data_norm.model[event],
                data_norm.model[event].replace(" ", ""),
                data_norm.acquisition_plane[event],
            )

            # Fetch k_tab
            c.execute(
                "SELECT k_patient_support "
                "FROM correction_table_and_pad_attenuation "
                "WHERE kvp_kv=? AND "
                "      filtration_added_mmcu=? AND "
                "      filtration_added_mmal=? AND "
                "      device_model IN (?, ?) AND "
                "      acquisition_plane=?",
                params,
            )

            k_tab[event] = c.fetchone()[0]
    finally:
        conn.commit()
        conn.close()

    return k_tab



import logging
from typing import Any, Dict, List

import numpy as np
import pandas as pd
from scipy.interpolate import CubicSpline
from tqdm import tqdm

from pyskindose import Phantom
from pyskindose import constants as c
from pyskindose.calculate_dose.add_correction_and_event_dose_to_output import (
    add_corrections_and_event_dose_to_output,
)
from pyskindose.calculate_dose.perform_calculations_for_new_geometries import (
    perform_calculations_for_new_geometries,
)

logger = logging.getLogger(__name__)


def calculate_irradiation_event_result(
    normalized_data: pd.DataFrame,
    event: int,
    total_events: int,
    new_geometry: List[bool],
    k_tab: List[float],
    hits: List[bool],
    patient: Phantom,
    table: Phantom,
    pad: Phantom,
    back_scatter_interpolation: List[CubicSpline],
    output: Dict[str, Any],
    corrections_db: str,
    table_hits: List[bool] = None,
    field_area: List[float] = None,
    k_isq: np.array = None,
    pbar: tqdm = None,
) -> Dict[str, Any]:
    """Conducts skin dose calculation.

    This function loops though all irradiation events in the the normalized data, and
    calculates the skin dose contribution from each event.

    Parameters
    ----------
    normalized_data : pd.DataFrame
        RDSR data, normalized for compliance with PySkinDose.
    event : starting event index
        Index of starting irradiation event
    total_events :
        Total number of irradiation events
    new_geometry : List[bool]
        A boolean list that specifies whether the irradiation geometry has changes since
        the preceding event. See the function check_new_geometry
    k_tab : List[float]
        List of table correction factors
    hits : List[bool]
        A boolean list that specifies (for a single event) the hit/miss status of each
        skin cell upon the patient phantom.
    patient : Phantom
        Patient skin surface phantom
    table : Phantom
        Patient support table phantom
    pad : Phantom
        Patient support pad phantom
    back_scatter_interpolation : List[CubicSpline]
        List of interpolation objects to used to estimate backscatter correction from
        the correction database
    output : Dict[str, Any]
        Dictionary containing outputs to store from the calculations. E.g. dose map and
        correction factors.
    corrections_db : str
        A string defining the path to the corrections SQLite db
    table_hits : List[bool], optional
        A boolean list that specfies (for each hit), if the bean passes through the
        patient support table, by default None
    field_area : List[float], optional
        X-ray field area in (cm^2) for each phantom skin cell that are hit by the X-ray
        beam, by default None
    k_isq : np.array, optional
        Inverse-square-law correction factors, by default None
    pbar : tqdm
        progress bar object

    Returns
    -------
    Dict[str, Any]
        Dictionary containing skin dose calculation results.

    """
    logger.debug(f"Calculating irradiation event {event + 1} out of {total_events}")

    hits, table_hits, field_area, k_isq = perform_calculations_for_new_geometries(
        normalized_data=normalized_data,
        event=event,
        new_geometry=new_geometry[event],
        patient=patient,
        table=table,
        pad=pad,
        hits=hits,
        table_hits=table_hits,
        field_area=field_area,
        k_isq=k_isq,
    )

    logger.debug("Saving event data")

    output[c.OUTPUT_KEY_HITS][event] = hits
    output[c.OUTPUT_KEY_KERMA][event] = normalized_data.K_IRP[event]
    output[c.OUTPUT_KEY_CORRECTION_INVERSE_SQUARE_LAW][event] = k_isq

    output = add_corrections_and_event_dose_to_output(
        normalized_data=normalized_data,
        event=event,
        hits=hits,
        table_hits=table_hits,
        patient=patient,
        back_scatter_interpolation=back_scatter_interpolation,
        field_area=field_area,
        k_tab=k_tab,
        output=output,
        corrections_db=corrections_db,
    )

    event += 1
    if event < total_events:

        pbar.update()

        output = calculate_irradiation_event_result(
            normalized_data=normalized_data,
            event=event,
            total_events=total_events,
            new_geometry=new_geometry,
            k_tab=k_tab,
            hits=hits,
            patient=patient,
            table=table,
            pad=pad,
            back_scatter_interpolation=back_scatter_interpolation,
            output=output,
            corrections_db=corrections_db,
            table_hits=table_hits,
            field_area=field_area,
            k_isq=k_isq,
            pbar=pbar,
        )

        if event == total_events - 1:
            pbar.update()
            pbar.refresh()
    return output



import logging
from typing import Any, List

import numpy as np
import pandas as pd

import pyskindose.constants as c

from .db_connect import db_connect
from .phantom_class import Phantom

logger = logging.getLogger(__name__)


def calculate_field_size(field_size_mode, data_parsed, data_norm):
    """Calculate X-ray field size at image recepter plane.

    Parameters
    ----------
    field_size_mode : str
        Choose either 'CFA' ('collimated field area) or 'ASD' (actual shutter
        distance).

        If field_size_mode = 'CFA', the field side in lateral- and
        longutudinal direction are set equal to the square root of the
        collimated field area. NOTE, this should only be used when actual
        shutter distances are unavailabe.

        If field_size_mode = 'ASD', the function calculates the field size
        by distance scaling the actual shutter distance to the detector plane

    data_parsed : pd.DataFrame
        Parsed RDSR data from all irradiation events in the RDSR input file,
        i.e. output of function rdsr_parser
    data_norm : pd.DataFrame
        RDSR data, normalized for compliance with PySkinDose.

    Returns
    -------
    FS_lat, FS_long : float
        Field size in lat- and long direction in cm at the detector plane.

    """
    # if collimated field are mode, set FS_lat = FS_long =
    # sqrt(collimate field area). NOTE: This should only be used when actual
    # shutter distances are unavailable.
    if field_size_mode == c.FIELD_SIZE_MODE_COLLIMATED_FIELD_AREA:

        # Field size in lat direction in metres
        FS_lat_m = np.sqrt(data_parsed.CollimatedFieldArea_m2)
        # Field size in long direction in metres
        FS_long_m = FS_lat_m

        # convert to cm
        FS_lat = round(convert_from_m_to_cm(FS_lat_m), 3)
        FS_long = round(convert_from_m_to_cm(FS_long_m), 3)

        return FS_lat, FS_long

    if field_size_mode == c.FIELD_SIZE_MODE_ACTUAL_SHUTTER_DISTANCE:

        # distance (from source) at which shutter distances are presented in
        # dicom tags
        d_shutter_in_dcm_cm = 100

        # field size in long direction at d = d_shutter_in_dcm_cm
        FS_long_at_d_shutter = convert_from_mm_to_cm(data_parsed.LeftShutter_mm + data_parsed.RightShutter_mm)
        # field size in lat direction at d = d_shutter_in_dcm_cm
        FS_lat_at_d_shutter = convert_from_mm_to_cm(data_parsed.TopShutter_mm + data_parsed.BottomShutter_mm)
        # scale factor to get field size at detector plane
        scale = data_norm.DSD / d_shutter_in_dcm_cm

        FS_lat = scale * FS_lat_at_d_shutter
        FS_long = scale * FS_long_at_d_shutter

        return FS_lat, FS_long


def position_patient_phantom_on_table(
    patient: Phantom,
    table: Phantom,
    pad: Phantom,
    pad_thickness: Any,
    patient_offset: List[int],
    patient_orientation: c.PATIENT_ORIENTATION_HEAD_FIRST_SUPINE,
) -> None:
    """Places the patient phantom upon the patient support table.

    In this function, the patient phantom is positioned to the starting
    position for the procedure. This is done by rotating and translating the
    patient so that the correct starting position is achieved.

    Parameters
    ----------
    patient : Phantom
        Patient phantom, either plane, cylinder or human.
    table : Phantom
        Table phantom to represent the patient support table
    pad : Phantom
        Pad phantom to represent the patient support pad
    pad_thickness: Any
        Patient support pad thickness
    patient_offset : List[int]
        Offsets the patient phantom from the centered along the head end of the
        table top, given as [Tx: <int>, "Ty": <int>, "Tz": <int>] in cm.
    patient_orientation : str
        patient orientation upon table. Choose between
        c.PATIENT_ORIENTATION_HEAD_FIRST_SUPINE and
        c.PATIENT_ORIENTATION_FEET_FIRST_SUPINE.

    """
    # if feet-first, rotate patient 180 degrees about y-axis
    if patient_orientation == c.PATIENT_ORIENTATION_FEET_FIRST_SUPINE:

        offset_to_rotation_center = abs(patient.r[:, 2].min()) / 2

        patient.translate(dr=[0, 0, +offset_to_rotation_center])
        patient.rotate(angles=[0, 180, 0])
        patient.translate(dr=[0, 0, -offset_to_rotation_center])

    # place the phantom on top of the support pad
    patient.translate(dr=[0, -pad_thickness, 0])
    # add user specified patient offset
    patient.translate(dr=patient_offset)

    # Save reference table position:
    for phantom in [table, pad, patient]:
        phantom.save_position()


def vector(start: np.array, stop: np.array, normalization=False) -> np.array:
    """Create a vector between two points in carthesian space.

    This function creates a simple vector between point <start> and point
    <stop> The function can also create a unit vector from <start>, in the
    direction to <stop>.

    Parameters
    ----------
    start : np.array
        Starting point of the vector
    stop : np.array
        Stopping point of the vector
    normalization : bool, optional
        Toggle normalization (the default is False, which implies no
        normalization)

    Returns
    -------
    np.array
        A vector from "start" to "stop", or if normalization=True, a unit
        vector from "start" in the direction towards "stop".

    """
    # Calculate vector from start to stop
    vec = stop - start

    # Normalize if requested
    if normalization:
        # Normalize vector
        mag = np.sqrt(vec.dot(vec))
        vec = vec / mag

    return vec


def scale_field_area(
    data_norm: pd.DataFrame, event: int, patient: Phantom, hits: List[bool], source: np.array
) -> List[float]:
    """Scale X-ray field area from image detector, to phantom skin cells.

    This function scales the X-ray field size from the point where it is stated
    in data_norm, i.e. at the image detector plane, to the plane at the phantom
    skin cell. This is the field size of interest since this area is required
    as input for k_med and k_bs correction factor calculations. This function
    conducts this scaling for all skin cells that are hit by the X-ray beam in
    a specific irradiation event.

    Parameters
    ----------
    data_norm : pd.DataFrame
        RDSR data, normalized for compliance with PySkinDose.
    event : int
        Irradiation event index.
    patient : Phantom
        Patient phantom, i.e. instance of class Phantom.
    hits : List[bool]
        A boolean list of the same length as the number of patient skin
        cells. True for all entrance skin cells that are hit by the beam for a
        specific irradiation event.
    source : np.array
        (x,y,z) coordinates to the X-ray source

    Returns
    -------
    List[float]
        X-ray field area in (cm^2) for each phantom skin cell that are hit by
        X-ray the beam

    """
    # Fetch reference distance for field size scaling,
    # i.e. distance source to detector
    d_ref = data_norm.DSD[event]

    cells = patient.r[hits]

    # Calculate distance scale factor
    scale_factor = [np.linalg.norm(cell - source) / d_ref for cell in cells]

    # Fetch field side lenth lateral and longitudinal at detector plane
    # Fetch field area at image detector plane
    field_area_ref = data_norm.FS_lat[event] * data_norm.FS_long[event]

    # Calculate field area at distance source to skin cell for all cells
    # that are hit by the beam.
    field_area = [round(field_area_ref * np.square(scale), 1) for scale in scale_factor]

    return field_area


def fetch_and_append_hvl(data_norm: pd.DataFrame, inherent_filtration: float, corrections_db: str) -> pd.DataFrame:
    """Add event HVL to RDSR event data from database.

    Parameters
    ----------
    data_norm : pd.DataFrame
        RDSR data, normalized for compliance with PySkinDose.
    inherent_filtration : float
        X-ray tube inherent filtration in mmAl.
    corrections_db : str
        A string defining the path to the corrections SQLite db

    Returns
    -------
    data_norm
        This function appends event specific HVL (mmAl) as a function of device
        model, kVp, and copper- and aluminum filtration to the normalized RDSR
        data in data_norm and returns the DataFrame with the HVL info appended.

    """
    # Open connection to database
    conn = db_connect(db_name=corrections_db)[0]

    # Fetch entire HVL table
    hvl_data = pd.read_sql_query("SELECT * FROM hvl_combined", conn)

    hvl = [
        float(
            hvl_data.loc[
                (round(hvl_data["kvp_kv"]) == round(data_norm.kVp[event]))
                & (round(hvl_data["filtration_inherent_mmal"], 1) == round(inherent_filtration, 1))
                & (hvl_data["filtration_added_mmcu"] == data_norm.filter_thickness_Cu[event])
                & (hvl_data["filtration_added_mmal"] == round(data_norm.filter_thickness_Al[event])),
                "hvl_mmal",
            ].iloc[0]
        )
        for event in range(len(data_norm))
    ]

    # Append HVL data to data_norm
    data_norm["HVL"] = hvl

    # close database connection
    conn.commit()
    conn.close()

    return data_norm


def check_new_geometry(data_norm: pd.DataFrame) -> List[bool]:
    """Check which events has unchanged geometry since the event before.

    This function is intended to calculate if new geometry parameters needs
    to be calculated, i.e., new beam, geometry positioning, field area and
    cell hit calculation.

    Parameters
    ----------
    data_norm : pd.DataFrame
        RDSR data, normalized for compliance with PySkinDose.

    Returns
    -------
    List[bool]
        List of booleans where True[event] means that the event has updated
        geometry since the preceding irradiation event.

    """
    logger.info("Checking which irradiation events contain changes in geometry" "compared to previous event")

    logger.debug("Listing all RDSR geometry parameters")
    geom_params = data_norm[["Tx", "Ty", "Tz", "FS_lat", "FS_long", "Ap1", "Ap2", "Ap3", "At1", "At2", "At3"]]

    logger.debug("Checking which irradiation events that does not have same" "parameters as previous")
    changed_geometry = [
        not geom_params.iloc[event].equals(geom_params.iloc[event - 1]) for event in range(1, len(geom_params))
    ]

    logger.debug("Insert True to the first event to indicate that it has a" "new geometry")

    changed_geometry.insert(0, True)

    return changed_geometry


class Triangle:
    """A class used to create triangles.

    This class creates a triangle from a set of three coordinates in 3D
    carthesian space. The purpose of this class is to use it to calculate if a
    3D segment intercepts the triangle.

    Attributes
    ----------
    p: np.array
        Carthesian 3D coordinates to the central vertex of the triangle
    p1: np.array
        Vector from p to first vertex
    p2: np.array
        Vector from p to second vertex
    n: np.array
        normal vector to the triangle, pointing upwards (negative y direction).

    """

    def __init__(self, p: np.array, p1: np.array, p2: np.array):
        """Initialize class attributes."""
        self.p = p
        self.p1 = vector(self.p, p1)
        self.p2 = vector(self.p, p2)
        n = np.cross(self.p1, self.p2)
        self.n = n / np.sqrt(n.dot(n))

    def check_intersection(self, start: np.array, stop: np.array) -> List[bool]:
        """Check if a 3D segment intercepts with the triangle.

        Check if a 3D segment intercepts with the triangle. For our purpose,
        the 3D segment represents an X-ray beam from the X-ray source to the
        phantom skin cell and the triangle represent parts of the patient
        support table. If the beam intercepts, table- and pad fluence
        correction should be conducted when calculating skin dose for that
        particular cell.

        Parameters
        ----------
        start : np.array
            Carthesian 3D coordinates to the starting point of the segment.
        stop : np.array
            Carthesian 3D coordinates to the end points of the segment. Note,
            can be several points, e.g, several skin cells.

        Returns
        -------
            List[bool]
            Boolean list which specifies whether each segment between start
            and each of coordinates in stop are intercepted by the triangle.

        """
        # Vector from source to central vertex
        # w = vector(start, self.p)
        w = self.p - start

        # List of unit vectors from start, to each of the coordinates in stop.
        v = ((stop - start).T / np.linalg.norm(stop - start, axis=stop.ndim - 1)).T

        # Distances from start to the plane of the triangle, in the direction
        # along the vector v.
        k = (np.dot(w, self.n)) / (np.dot(v, self.n))
        # Vector from origin to beam-table interceptions.
        i = start + (k * v.T).T

        # Vector from central vertex p to i
        p_i = i - self.p

        d = np.square(np.dot(self.p1, self.p2)) - np.dot(self.p1, self.p1) * np.dot(self.p2, self.p2)

        d1 = (np.dot(self.p1, self.p2) * np.dot(p_i, self.p2) - np.dot(self.p2, self.p2) * np.dot(p_i, self.p1)) / d

        d2 = (np.dot(self.p1, self.p2) * np.dot(p_i, self.p1) - np.dot(self.p1, self.p1) * np.dot(p_i, self.p2)) / d

        # Now we have p_i = d1/d * p1 + d2/d * p2, thus,
        # if 0 <= d1/d <= 1, and 0 <= d2/d <= 1, and d1 + d2 <= 1, the beam
        # intercepts the triangle.
        hits = np.array([d1 >= 0, d1 <= 1, d2 >= 0, d2 <= 1, d1 + d2 <= 1]).all(axis=0)

        return hits.tolist()


def check_table_hits(source: np.array, table: Phantom, beam, cells: np.array) -> List[bool]:
    """Check which skin cells are blocket by the patient support table.

    This fuctions creates two triangles covering the entire surface of the
    patient support table, and checks if the skin cells are blocked by the
    table. This is conducted in order to be able to append table and pad
    correction factor k_(T+P) when required.

    Parameters
    ----------
    source : np.array
        Carthesian 3D coordinates to the X-ray souce
    table : Phantom
        Patient support table, i.e., instance of class phantom with
        phantom_type="table"
    beam : Beam
         X-ray beam, i.e., instance of class Beam.
    cells : np.array
        List of skin cells to be controlled if the patient support table and
        pad blocks the beam before it reached the them.

    Returns
    -------
    List[bool]
        Boolean list of the statuses of each skin cell. True if the path from
        X-ray source to skin cell is blocked by the table (any of the two
        triangles), else false. Start points above triangle returns False,
        to not include hits where the table does not block the beam.

    """
    # Create triangles:

    # Define edges of table (see illustration in project documentation)
    a = table.r[0 + 4, :]
    a1 = table.r[3 + 4, :]
    a2 = table.r[1 + 4, :]

    b = table.r[2 + 4, :]
    b1 = table.r[1 + 4, :]
    b2 = table.r[3 + 4, :]

    # triangle spanning the "top right" part of the support table
    # (viewed from above)
    triangle_b_l = Triangle(p=a, p1=a1, p2=a2)
    # triangle spanning the "bottom left" part of the support table
    # (viewed from above)
    triangle_t_r = Triangle(p=b, p1=b1, p2=b2)

    # If over-table irradiation, return false for all points in cells
    if np.dot(np.array([0, 0, 0]) - beam.r[0, :], triangle_b_l.n) < 0:
        if cells.ndim == 1:
            return [False]
        return [False] * cells.shape[0]

    # Check if beam vertices hits table on either of the triangles
    hit_t_r = triangle_t_r.check_intersection(start=source, stop=beam.r[1:, :])
    hit_b_l = triangle_b_l.check_intersection(start=source, stop=beam.r[1:, :])

    # If all four beam verices hits the table, all cells are blocket by the
    # table, and all cells should be corrected for table and pad attenuation.
    if sum(hit_t_r + hit_b_l) == 4:
        if cells.ndim == 1:
            return [True]
        return [True] * cells.shape[0]

    # Else, check individually for all skin cells that are hit by the beam
    hit_t_r = triangle_t_r.check_intersection(start=source, stop=cells)
    hit_b_l = triangle_b_l.check_intersection(start=source, stop=cells)

    hits = np.asarray([False] * len(cells))
    # save results
    hits[hit_t_r] = True
    hits[hit_b_l] = True

    return hits.tolist()


def convert_from_mm_to_cm(val_in_mm: float) -> float:
    """Convert a length from millimeters to centimeters.

    Parameters
    ----------
    val_in_mm : float
        A length in mm

    Returns
    -------
    float
        The same length in cm

    """
    val_in_cm = val_in_mm / 10.0

    return val_in_cm


def convert_from_m_to_cm(val_in_m: float) -> float:
    """Convert a length from centimeters to millimeters.

    Parameters
    ----------
    val_in_m : float
        A length in m

    Returns
    -------
    float
        The same length in cm

    """
    val_in_cm = val_in_m * 100.0

    return val_in_cm



import pandas as pd

from pyskindose import constants as c
from pyskindose import position_patient_phantom_on_table
from pyskindose.phantom_class import Phantom
from pyskindose.plotting.plot_geometry import plot_geometry
from pyskindose.settings import PyskindoseSettings


def create_geometry_plot(
    normalized_data: pd.DataFrame,
    table: Phantom,
    pad: Phantom,
    settings: PyskindoseSettings,
) -> None:
    """Create any of the geometry plots available in PySkinDose.

    Parameters
    ----------
    normalized_data : pd.DataFrame
        RDSR data, normalized for compliance with PySkinDose
    table : Phantom
        Patient support table phantom
    pad : Phantom
        Patient support pad phantom
    settings : PyskindoseSettings
        Settings class for PySkinDose

    """
    if settings.mode not in [
        c.MODE_PLOT_SETUP,
        c.MODE_PLOT_EVENT,
        c.MODE_PLOT_PROCEDURE,
    ]:
        return

    # override dense mathematical phantom in .html plotting
    if settings.phantom.model == c.PHANTOM_MODEL_PLANE:
        settings.phantom.dimension.plane_resolution = c.RESOLUTION_SPARSE
    elif settings.phantom.model == c.PHANTOM_MODEL_CYLINDER:
        settings.phantom.dimension.cylinder_resolution = c.RESOLUTION_SPARSE

    patient = Phantom(
        phantom_model=settings.phantom.model,
        phantom_dim=settings.phantom.dimension,
        human_mesh=(
            settings.phantom.human_mesh
            if isinstance(settings.phantom.human_mesh, tuple)
            else f"{settings.phantom.human_mesh}{c.PHANTOM_HUMAN_MESH_SPARSE_MODEL_ENDING if settings.mode == c.MODE_PLOT_PROCEDURE and settings.phantom.model == c.PHANTOM_MODEL_HUMAN else ''}"
        ),  # override dense .stl phantoms in plot_procedure .html plotting
    )

    # position objects in starting position
    position_patient_phantom_on_table(
        patient=patient,
        table=table,
        pad=pad,
        pad_thickness=settings.phantom.dimension.pad_thickness,
        patient_offset=[
            settings.phantom.patient_offset.d_lon,
            settings.phantom.patient_offset.d_ver,
            settings.phantom.patient_offset.d_lat,
        ],
        patient_orientation=settings.phantom.patient_orientation,
    )

    plot_geometry(
        patient=patient,
        table=table,
        pad=pad,
        data_norm=normalized_data,
        mode=settings.mode,
        event=settings.plot.plot_event_index,
        dark_mode=settings.plot.dark_mode,
        notebook_mode=settings.plot.notebook_mode,
        include_patient=len(normalized_data) <= settings.plot.max_events_for_patient_inclusion,
    )
